/*
 * PCI code for DDB5477.
 *
 * Copyright (C) 2001 MontaVista Software Inc.
 * Author: Jun Sun, jsun@mvista.com or jsun@junsun.net
 *
 * Copyright (C) 2004 by Ralf Baechle (ralf@linux-mips.org)
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/pci.h>

#include <asm/bootinfo.h>
#include <asm/debug.h>

#include <asm/byteorder.h>
//#include <asm/pci_channel.h>

#include <asm/tcube.h>

static struct resource extpci_io_resource = {
	"ext pci IO space", 
	0,
	0x007FFFFF,
	IORESOURCE_IO};

static struct resource extpci_mem_resource = {
	"ext pci memory space", 
	0x10000000,
	0x17FFFFFF,
	IORESOURCE_MEM};

static struct resource iopci_io_resource = {
	"io pci IO space", 
	0x00800000,
	0x00FFFFFF,
	IORESOURCE_IO};

static struct resource iopci_mem_resource = {
	"io pci memory space", 
	0x19000000,
	0x197FFFFF,
	IORESOURCE_MEM};

extern struct pci_ops ddb5477_ext_pci_ops;
extern struct pci_ops ddb5477_io_pci_ops;

struct pci_controller ddb5477_ext_controller = {
	.pci_ops	= &ddb5477_ext_pci_ops,
	.io_resource	= &extpci_io_resource,
	.mem_resource	= &extpci_mem_resource
};

struct pci_controller ddb5477_io_controller = {
	.pci_ops	= &ddb5477_io_pci_ops,
	.io_resource	= &iopci_io_resource,
	.mem_resource	= &iopci_mem_resource
};

#define NUM_5701_IRQS 	  32
#define NUM_5701_EPCI_IRQ 8

static void tcube_pci_device_fixup(void)
{
	volatile unsigned long i;
	static int _ini = 0;
	
	if(_ini==0)_ini=1;
	else       return;

	printk("tcube_pci_device_fixup().in\n");

//	printk("fixup-IDE...\n");
//	// VR5701-IDE-RESET CLEAR(KOSE)
//	*(volatile unsigned char *)0xb8801030 = 6 ; //pci.io.space
//	asm("sync");for(i=0;i<0x1000000;i++);

	printk("fixup-SM722...\n");
}


int __init pcibios_map_irq(struct pci_dev *dev, u8 slot, u8 pin)
{
	int slot_num;
//	unsigned char *slot_irq_map;
//	unsigned char irq;
    int k;

	slot_num = PCI_SLOT(dev->devfn);

		if( dev->bus->number == 0 ){ // EPCI
			switch( slot_num ){
			case 25-11:			// USB    INTC#
				k = NUM_5701_IRQS + 2 ; break ;
			case 26-11:                     // LAN    INTB#
				k = NUM_5701_IRQS + 1 ; break ;
			case 27-11:                     // VIDEO  INTA#
				k = NUM_5701_IRQS + 0 ; break ;
			}
		}else{                       // IPCI
			switch( slot_num ){
			case 29-11:			// AC97    INTC#
				k = NUM_5701_IRQS + NUM_5701_EPCI_IRQ + 2 ; break ;
			case 30-11:                     // IDE     INTB#
				k = NUM_5701_IRQS + NUM_5701_EPCI_IRQ + 1 ; break ;
			case 31-11:                     // USB     INTA#
				k = NUM_5701_IRQS + NUM_5701_EPCI_IRQ + 0 ; break ;
			}
		}
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, k);
		dev->irq = k+8;

	tcube_pci_device_fixup();
	return dev->irq;
}

void ddb_pci_reset_bus(void)
{
	u32 temp;
volatile unsigned long i;

printk("KOSE ddb_pci_reset_bus\n");

	/*
	 * I am not sure about the "official" procedure, the following
	 * steps work as far as I know:
	 * We first set PCI cold reset bit (bit 31) in PCICTRL-H.
	 * Then we clear the PCI warm reset bit (bit 30) to 0 in PCICTRL-H.
	 * The same is true for both PCI channels.
	 */
	temp = ddb_in32(EPCI_CTRLH);
	temp |= 0x80000000;
	ddb_out32(EPCI_CTRLH, temp);
for(i=0;i<1000;i++);
	temp &= ~0xc0000000;
	ddb_out32(EPCI_CTRLH, temp);

	temp = ddb_in32(IPCI_CTRLH);
	temp |= 0x80000000;
	ddb_out32(IPCI_CTRLH, temp);
for(i=0;i<1000;i++);
	temp &= ~0xc0000000;
	ddb_out32(IPCI_CTRLH, temp);
}
