#define PCIX_BRIDGE

#include <linux/config.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/types.h>
#include <linux/ptrace.h>

#include <asm/system.h>
#include <asm/mipsregs.h>
#include <asm/debug.h>
#include <asm/tcube.h>

/*
 * IRQ mapping
 *
 *  0-7: 8 CPU interrupts
 *	0 -	software interrupt 0
 *	1 -	software interrupt 1
 *	2 -	most Vrc5477 interrupts are routed to this pin
 *	3 -	(optional) some other interrupts routed to this pin for debugg
 *	4 -	not used
 *	5 -	not used
 *	6 -	not used
 *	7 -	cpu timer (used by default)
 *
 */

#define ACTIVE_LOW		1
#define ACTIVE_HIGH		0

#define LEVEL_SENSE		2
#define EDGE_TRIGGER		0

#define INTA			0
#define INTB			1
#define INTC			2
#define INTD			3
#define INTE			4

extern void tcube_irq_init(u32 base);
extern void mips_cpu_irq_init(u32 base);
extern asmlinkage void tcube_handle_int(void);
extern void vr5701_irq_init(u32 irq_base);

extern void ll_vr5701_irq_route(int vr5701_irq, int ip);
extern void ll_vr5701_irq_enable(int vr5701_irq);
extern void ll_vr5701_irq_disable(int vr5701_irq, int ack);

void __init arch_init_irq(void)
//void tcube_irq_setup(void)
{
	db_run(printk("ddb5477_irq_setup invoked.\n"));

	/* by default, we disable all interrupts and route all vr5701 
	 * interrupts to pin 0 (irq 2) */
	ddb_out32(INT_ROUTE0, 0);
	ddb_out32(INT_ROUTE1, 0);
	ddb_out32(INT_ROUTE2, 0);
	ddb_out32(INT_ROUTE3, 0);

	clear_c0_status(0xff00);
	set_c0_status(0x0400);

	/* setup PCI interrupt attributes */
//	set_pci_int_attr(PCI0, INTA, ACTIVE_LOW, LEVEL_SENSE);
//	set_pci_int_attr(PCI0, INTB, ACTIVE_LOW, LEVEL_SENSE);
//	set_pci_int_attr(PCI0, INTC, ACTIVE_LOW, LEVEL_SENSE);
//	set_pci_int_attr(PCI0, INTD, ACTIVE_LOW, LEVEL_SENSE);
//	set_pci_int_attr(PCI0, INTE, ACTIVE_LOW, LEVEL_SENSE);

//	set_pci_int_attr(PCI1, INTA, ACTIVE_LOW, LEVEL_SENSE);
//	set_pci_int_attr(PCI1, INTB, ACTIVE_LOW, LEVEL_SENSE);
//	set_pci_int_attr(PCI1, INTC, ACTIVE_LOW, LEVEL_SENSE);
//	set_pci_int_attr(PCI1, INTD, ACTIVE_LOW, LEVEL_SENSE);
//	set_pci_int_attr(PCI1, INTE, ACTIVE_LOW, LEVEL_SENSE);

#ifdef KOSE___________
	/* 
	 * for debugging purpose, we enable several error interrupts
	 * and route them to pin 1. (IP3) 
	 */
	/* cpu parity check - 0 */
	ll_vr5701_irq_route(0, 1); ll_vr5701_irq_enable(0);
	/* cpu no-target decode - 1 */
	ll_vr5701_irq_route(1, 1); ll_vr5701_irq_enable(1);
	/* local bus read time-out - 7 */
	ll_vr5701_irq_route(7, 1); ll_vr5701_irq_enable(7);
	/* PCI SERR# - 14 */
	ll_vr5701_irq_route(14, 1); ll_vr5701_irq_enable(14);
	/* PCI internal error - 15 */
	ll_vr5701_irq_route(15, 1); ll_vr5701_irq_enable(15);
	/* IOPCI SERR# - 30 */
	ll_vr5701_irq_route(30, 1); ll_vr5701_irq_enable(30);
	/* IOPCI internal error - 31 */
	ll_vr5701_irq_route(31, 1); ll_vr5701_irq_enable(31);
#endif
	ll_vr5701_irq_route(24, 1); ll_vr5701_irq_enable(24);	// tbus err
	ll_vr5701_irq_route(25, 1); ll_vr5701_irq_enable(25);   // local err
	ll_vr5701_irq_route(28, 1); ll_vr5701_irq_enable(28);   // epci ierr
	ll_vr5701_irq_route(29, 1); ll_vr5701_irq_enable(29);   // epci serr
	ll_vr5701_irq_route(30, 1); ll_vr5701_irq_enable(30);   // ipci ierr
	ll_vr5701_irq_route(31, 1); ll_vr5701_irq_enable(31);   // ipci serr

	/* init all controllers */
	mips_cpu_irq_init(0);
	vr5701_irq_init(8);

	/* hook up the first-level interrupt handler */
	set_except_vector(0, tcube_handle_int);
}

#define NUM_5701_IRQS 	  32
#define NUM_5701_EPCI_IRQ 8
/*
 * the first level int-handler will jump here if it is a vr7701 irq
 */

asmlinkage void
tcube_irq_dispatch(struct pt_regs *regs)
{
	extern unsigned int do_IRQ(int irq, struct pt_regs *regs);

	u32 intStatus;
	u32 bitmask;
	u32 i;
	int pci_flag = 0;
	int ipci_flag = 0;

	db_assert(ddb_in32(INT2_STAT) == 0);
	db_assert(ddb_in32(INT3_STAT) == 0);
	db_assert(ddb_in32(INT4_STAT) == 0);
	db_assert(ddb_in32(NMI_STAT) == 0);

	if (ddb_in32(INT1_STAT) != 0) {
#if defined(CONFIG_DEBUG)
//KOSE.DEL	vrc5477_show_int_regs();
#endif
		panic("error interrupt has happened.");
	}

	intStatus = ddb_in32(INT0_STAT);
	for (i=0, bitmask=1; i<= NUM_5701_IRQS; bitmask <<=1, i++) {
		/* do we need to "and" with the int mask? */
		if (intStatus & bitmask) {
			switch( i ){
			case 6: pci_flag = 1 ;  break ;
			case 7: ipci_flag = 1 ; break ;
			default : do_IRQ(8 + i, regs); break ;
			}
		}
	}
	if( pci_flag ){
		intStatus = ddb_in32(EPCI_INTS);
		for (i=0, bitmask=1; i< 4 ; bitmask<<=1, i++) {
			if (intStatus & bitmask) {
//if(i==2)printk("[KOSE USB]");
				do_IRQ(8 + NUM_5701_IRQS + i, regs);
			}
		}
	}
	if( ipci_flag ){
		intStatus = ddb_in32(IPCI_INTS);
		for (i=0, bitmask=1; i< 4 ; bitmask<<=1, i++) {
			if (intStatus & bitmask) {
				do_IRQ(8 + NUM_5701_IRQS + NUM_5701_EPCI_IRQ + i, regs);
			}
		}
	}
	
}
