
/*
 * Vr5701 defines 32 IRQs.
 *
 */
#include <linux/interrupt.h>
#include <linux/types.h>
#include <linux/ptrace.h>

#include <asm/debug.h>
#include <asm/tcube.h>

static int vr5701_irq_base = -1;

#define NUM_5701_IRQ  32
#define NUM_5701_EPCI_IRQ  8
#define NUM_5701_IPCI_IRQ  8

void ll_vr5701_irq_route(int vr5701_irq, int ip);
void ll_vr5701_irq_enable(int vr5701_irq);
void ll_vr5701_irq_disable(int vr5701_irq, int ack);

static void 
vr5701_irq_enable(unsigned int irq)
{
	db_assert(vr5701_irq_base != -1);
	db_assert(irq >= vr5701_irq_base);
	db_assert(irq < vr5701_irq_base + NUM_5701_IRQ + NUM_5701_EPCI_IRQ + NUM_5701_IPCI_IRQ);

	ll_vr5701_irq_enable(irq - vr5701_irq_base);
}

static void 
vr5701_irq_disable(unsigned int irq)
{
	db_assert(vr5701_irq_base != -1);
	db_assert(irq >= vr5701_irq_base);
	db_assert(irq < vr5701_irq_base + NUM_5701_IRQ + NUM_5701_EPCI_IRQ + NUM_5701_IPCI_IRQ);

	ll_vr5701_irq_disable(irq - vr5701_irq_base, 0 );
}

static unsigned int vr5701_irq_startup(unsigned int irq)
{
	vr5701_irq_enable(irq);
	return 0;
}

#define	vr5701_irq_shutdown	vr5701_irq_disable

static void
vr5701_irq_ack(unsigned int irq)
{
	db_assert(vr5701_irq_base != -1);
	db_assert(irq >= vr5701_irq_base);
	db_assert(irq < vr5701_irq_base + NUM_5701_IRQ + NUM_5701_EPCI_IRQ + NUM_5701_IPCI_IRQ );

	/* clear the interrupt bit */
	/* some irqs require the driver to clear the sources */
	if( irq < vr5701_irq_base + NUM_5701_IRQ ){
		ddb_out32(INT_CLR, 1 << (irq - vr5701_irq_base));
	}else if( irq < vr5701_irq_base + NUM_5701_IRQ + NUM_5701_EPCI_IRQ ){
		ddb_out32(INT_CLR, 1 << 6 );		// EPCI
	}else{
		ddb_out32(INT_CLR, 1 << 7 );		// IPCI
	}
	/* disable interrupt - some handler will re-enable the irq
	 * and if the interrupt is leveled, we will have infinite loop
	 */
	ll_vr5701_irq_disable(irq - vr5701_irq_base, 1 );
}

static void
vr5701_irq_end(unsigned int irq)
{
	db_assert(vr5701_irq_base != -1);
	db_assert(irq >= vr5701_irq_base);
	db_assert(irq < vr5701_irq_base + NUM_5701_IRQ + NUM_5701_EPCI_IRQ + NUM_5701_IPCI_IRQ);

	if(!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS))){
		ll_vr5701_irq_enable( irq - vr5701_irq_base);
	}
}

hw_irq_controller vr5701_irq_controller = {
	"vr5701_irq",
	vr5701_irq_startup,
	vr5701_irq_shutdown,
	vr5701_irq_enable,
	vr5701_irq_disable,
	vr5701_irq_ack,
	vr5701_irq_end,
	NULL			/* no affinity stuff for UP */
};

void 
vr5701_irq_init(u32 irq_base)
{
	extern irq_desc_t irq_desc[];
	u32 i;

#if 0
	// --------------------------
	// KOSE PCIX:=LEVEL triger
	// --------------------------
	ddb_out32(VR5701_LEVEL_EDGE, 0xffff);   /* edge trigger */
#endif
	/* printk("vr5701_irq_init(%d)\n",irq_base); */
	for (i= irq_base; i< irq_base + NUM_5701_IRQ + NUM_5701_EPCI_IRQ + NUM_5701_IPCI_IRQ ; i++) {
		irq_desc[i].status = IRQ_DISABLED;
		irq_desc[i].action = NULL;
		irq_desc[i].depth = 1;
		irq_desc[i].handler = &vr5701_irq_controller;
	}
	vr5701_irq_base = irq_base;
}


int vr5701_irq_to_irq(int irq)
{
	db_assert(irq >= 0);
	db_assert(irq < NUM_5701_IRQ + NUM_5701_EPCI_IRQ + NUM_5701_IPCI_IRQ);

	return irq + vr5701_irq_base;
}

void ll_vr5701_irq_route(int vr5701_irq, int ip)
{
	u32 reg_value;
	u32 reg_bitmask;
	u32 reg_index;

	if( vr5701_irq >= NUM_5701_IRQ && vr5701_irq < NUM_5701_IRQ + NUM_5701_EPCI_IRQ)
		vr5701_irq = 6;
	else if( vr5701_irq >= NUM_5701_IRQ + NUM_5701_EPCI_IRQ && vr5701_irq < NUM_5701_IRQ + NUM_5701_EPCI_IRQ + NUM_5701_IPCI_IRQ)
		vr5701_irq = 7;

	db_assert(vr5701_irq >= 0);
	db_assert(vr5701_irq < NUM_5701_IRQ);
	db_assert(ip >= 0);
	db_assert((ip < 5) || (ip == 6));

	reg_index = INT_ROUTE0 + vr5701_irq/8*4;
	reg_value = ddb_in32(reg_index);
	reg_bitmask = 7 << (vr5701_irq % 8 * 4);
	reg_value &= ~reg_bitmask;
	reg_value |= ip << (vr5701_irq % 8 * 4);
	ddb_out32(reg_index, reg_value);
}

void ll_vr5701_irq_enable(int vr5701_irq)
{
	u16 reg_value;
	int tmp_irq;
	u32 reg_bitmask;

	// EPCI
	if( vr5701_irq >= NUM_5701_IRQ && vr5701_irq < NUM_5701_IRQ + NUM_5701_EPCI_IRQ) {	
		tmp_irq = vr5701_irq ;
		vr5701_irq = 6 ;
	// IPCI
	}else if( vr5701_irq >= NUM_5701_IRQ + NUM_5701_EPCI_IRQ && vr5701_irq < NUM_5701_IRQ + NUM_5701_EPCI_IRQ + NUM_5701_IPCI_IRQ) {	
		tmp_irq = vr5701_irq ;
		vr5701_irq = 7 ;
	}

	db_assert(vr5701_irq >= 0);
	db_assert(vr5701_irq < NUM_5701_IRQ);
	reg_value = ddb_in32(INT_MASK);
	ddb_out32(INT_MASK, reg_value | (1 << vr5701_irq));

	if( vr5701_irq == 6 ){			// EPCI
		reg_value = ddb_in32(EPCI_INTM);
		reg_bitmask = 1 << (tmp_irq - NUM_5701_IRQ );
		db_assert((reg_value & reg_bitmask) == 0);
		ddb_out32(EPCI_INTM, reg_value | reg_bitmask);
	}else if( vr5701_irq == 7 ){		// IPCI
		reg_value = ddb_in32(IPCI_INTM);
		reg_bitmask = 1 << (tmp_irq - NUM_5701_IRQ - NUM_5701_EPCI_IRQ );
		db_assert((reg_value & reg_bitmask) == 0);
		ddb_out32(IPCI_INTM, reg_value | reg_bitmask);
	}
}

//static u16 udummy;
void ll_vr5701_irq_disable(int vr5701_irq, int ack)
{
	u16 reg_value;
	u32 udummy;
	u32 reg_bitmask;

	if( !ack ) return ;

	if( vr5701_irq >= NUM_5701_IRQ && vr5701_irq < NUM_5701_IRQ + NUM_5701_EPCI_IRQ)
		goto LLL_EPCI ;
	else if( vr5701_irq >= NUM_5701_IRQ + NUM_5701_EPCI_IRQ && vr5701_irq < NUM_5701_IRQ + NUM_5701_EPCI_IRQ + NUM_5701_IPCI_IRQ)
		goto LLL_IPCI ;

	db_assert(vr5701_irq >= 0);
	db_assert(vr5701_irq < NUM_5701_IRQ);
	reg_value = ddb_in32(INT_MASK);
	ddb_out32(INT_MASK, reg_value & ~(1 << vr5701_irq));
	udummy    = ddb_in32(INT_MASK);
        return ;

LLL_EPCI:;
	reg_value = ddb_in32(EPCI_INTM);
	reg_bitmask = 1 << (vr5701_irq - NUM_5701_IRQ );
	db_assert((reg_value & reg_bitmask) != 0);
	ddb_out32(EPCI_INTM, reg_value & ~reg_bitmask);
	return ;

LLL_IPCI:;
	reg_value = ddb_in32(IPCI_INTM);
	reg_bitmask = 1 << (vr5701_irq - NUM_5701_IRQ - NUM_5701_EPCI_IRQ );
	db_assert((reg_value & reg_bitmask) != 0);
	ddb_out32(IPCI_INTM, reg_value & ~reg_bitmask);
	return ;


}
