/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * SNI specific PCI support for RM200/RM300.
 *
 * Copyright (C) 1997 - 2000 Ralf Baechle
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <asm/byteorder.h>
#include <asm/pci_channel.h>

#include <asm/tcube.h>

static struct resource extpci_io_resource = {
	"ext pci IO space", 
	0,
	0x007FFFFF,
	IORESOURCE_IO};

static struct resource extpci_mem_resource = {
	"ext pci memory space", 
	0x10000000,
	0x17FFFFFF,
	IORESOURCE_MEM};

static struct resource iopci_io_resource = {
	"io pci IO space", 
	0x00800000,
	0x00FFFFFF,
	IORESOURCE_IO};

static struct resource iopci_mem_resource = {
	"io pci memory space", 
	0x19000000,
	0x197FFFFF,
	IORESOURCE_MEM};

extern struct pci_ops ddb5477_ext_pci_ops;
extern struct pci_ops ddb5477_io_pci_ops;

struct pci_channel mips_pci_channels[] = {
	{ &ddb5477_ext_pci_ops, &extpci_io_resource, &extpci_mem_resource },
	{ &ddb5477_io_pci_ops, &iopci_io_resource, &iopci_mem_resource },
	{ NULL, NULL, NULL}
};

void __init pcibios_fixup(void)
{
	// --------------------------------------------
	// Okamura module  >>> move to board init()
	// --------------------------------------------
}


#define NUM_5701_IRQS 	  32
#define NUM_5701_EPCI_IRQ 8

extern int vr5701_irq_to_irq(int irq);

static void tcube_pci_device_fixup(void)
{
	volatile unsigned long i;

	printk("tcube_pci_device_fixup().in\n");

	printk("fixup-IDE...\n");
	// VR5701-IDE-RESET CLEAR(KOSE)
	*(volatile unsigned char *)0xb8801030 = 6 ; //pci.io.space
	asm("sync");for(i=0;i<0x1000000;i++);

	printk("fixup-SM722...\n");
}


void __init pcibios_fixup_irqs(void)
{
        struct pci_dev *dev;
        int slot_num;
        int k;

printk("KOSE:pcibios_fixup_irqs.in\n");
	pci_for_each_dev(dev) {
		slot_num = PCI_SLOT(dev->devfn);
printk("slot=%d, bus=%d\n", slot_num, dev->bus->number );
		if( dev->bus->number == 0 ){ // EPCI
			switch( slot_num ){
			case 25-11:			// USB    INTC#
				k = NUM_5701_IRQS + 2 ; break ;
			case 26-11:                     // LAN    INTB#
				k = NUM_5701_IRQS + 1 ; break ;
			case 27-11:                     // VIDEO  INTA#
				k = NUM_5701_IRQS + 0 ; break ;
			}
		}else{                       // IPCI
			switch( slot_num ){
			case 29-11:			// AC97    INTC#
				k = NUM_5701_IRQS + NUM_5701_EPCI_IRQ + 2 ; break ;
			case 30-11:                     // IDE     INTB#
				k = NUM_5701_IRQS + NUM_5701_EPCI_IRQ + 1 ; break ;
			case 31-11:                     // USB     INTA#
				k = NUM_5701_IRQS + NUM_5701_EPCI_IRQ + 0 ; break ;
			}
		}
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, k);
		dev->irq = vr5701_irq_to_irq(k);
	}
	tcube_pci_device_fixup();
}

#define IO_MEM_LOGICAL_START   0x3e000000
#define IO_MEM_LOGICAL_END     0x3fefffff

#define IO_PORT_LOGICAL_START  0x3ff00000
#define IO_PORT_LOGICAL_END    0x3fffffff


#define IO_MEM_VIRTUAL_OFFSET  0xb0000000
#define IO_PORT_VIRTUAL_OFFSET 0xb0000000

#define ONE_MEG   (1024 * 1024)

void __init pci_setup(void)
{
#ifdef KOSE_______________
   printk("pci_setup() %s:%d\n",__FILE__,__LINE__);

   iomem_resource.start =  0x20000000 ;
   iomem_resource.end =    0x20FFFFFF ;
   ioport_resource.start = 0x0100 ;
   ioport_resource.end =   0xFFFF ;

   // map the PCI address space
   // global map - all levels & processes can access
   // except that the range is outside user space
   // parameters: lo0, lo1, hi, pagemask
   // lo indicates physical page, hi indicates virtual address

   printk("Allocating PCI a virtual space on kseg3\n");

   add_wired_entry( 
      ( 0x120000 << 6 ) | 0x17, // ENTLO0 MAPPING 0x1_2000_0000 (16MB)
      ( 0x121000 << 6 ) | 0x17, // ENTLO1 MAPPING 0x1_2100_0000 (16MB)
      ( 0xe0000000 ),
      PM_16M
   );

   // setup pcix bus as host bridge
//   io_out32(VR7701_PCIX_CMDSTAT, 0x00000006); // BusMaster, Memory = Enable 
//   io_out32(VR7701_PCIX_LATENCY, 0x00000200); // LATENCY = 0x20
//   io_out8(VR7701_PCIX_LATENCY_TIMER, 0x40);
   
//   io_out32(VR7701_PCIX_MMSEGL,  0x00000003); // offset = 0x00, no swap, no prefetchable
//   io_out32(VR7701_PCIX_MMSEGH,  0xffffffff); // offset = 0xffffffff
//   io_out32(VR7701_PCIX_TMMSK0,  0x1fff0001); // size = 512MB, enable
//   io_out32(VR7701_PCIX_TMMSK1,  0xffff0000); // size = 512MB, disable
//   io_out32(VR7701_PCIX_BAR0L,   0x0000000c); // pcix window base addr = 0 
//   io_out32(VR7701_PCIX_BAR0H,   0x00000000); // pcix window base addr = 0 
//   io_out32(VR7701_PCIX_BAR1L,   0x2000000c); // pcix window base addr = 0x20000000 
//   io_out32(VR7701_PCIX_BAR1H,   0x00000000); // pcix window base addr = 0 
//   io_out32(VR7701_PCIX_TMSEG0,  0x00000000); // oncip bus base addr = 0 
//   io_out32(VR7701_PCIX_TMSEG1,  0x00000000); // oncip bus base addr = 0 
//   io_out32(VR7701_PCIX_CMD,     0x003e0007); // max split = 4, max burst = 4096, ero = 1
//   io_out32(VR7701_PCIX_STAT,    0x00030008); // 133MHz, 64bit dev, dev = 1, func = 0
// io_out32(VR7701_PCIX_INTMASK, 0x00000000); // Disable all INT
#ifndef PCIX_BRIDGE
//   io_out32(VR7701_PCIX_INTMASK, 0x0000000E); // non_mask:=INTB,INTC,INTD, mask:=INTA
#else
//   io_out32(VR7701_PCIX_INTMASK, 0x0000000F); // non_mask:=INTA,INTB,INTC,INTD
#endif
//   io_out32(VR7701_PCIX_ERRPMASK,0x00000001); // Disable all pci error except master abort
//   io_out32(VR7701_PCIX_ERRPSTAT,0x000307ff); // Disable all pci error except master abort
//   io_out32(VR7701_PCIX_CONTROL, 0x80000020); // Disable all pci error except master abort
   
//   printk("VR7701_PCIX_LATENCY=%08lX\n", io_in32(VR7701_PCIX_LATENCY ));
//   printk("VR7701_PCIX_CONTROL=%08lX\n", io_in32(VR7701_PCIX_CONTROL ));

   printk("PCIX: finish initialization\n");
   
//   io_out32(VR7701_PCIX_RESET,   0x00000001); // CLEAR reset output
   printk("PCIX: clear reset \n");
#endif
}

void __init ddb_pci_reset_bus(void)
{
	u32 temp;
volatile unsigned long i;
	/*
	 * I am not sure about the "official" procedure, the following
	 * steps work as far as I know:
	 * We first set PCI cold reset bit (bit 31) in PCICTRL-H.
	 * Then we clear the PCI warm reset bit (bit 30) to 0 in PCICTRL-H.
	 * The same is true for both PCI channels.
	 */
	temp = ddb_in32(EPCI_CTRLH);
	temp |= 0x80000000;
	ddb_out32(EPCI_CTRLH, temp);
for(i=0;i<1000;i++);
	temp &= ~0xc0000000;
	ddb_out32(EPCI_CTRLH, temp);

	temp = ddb_in32(IPCI_CTRLH);
	temp |= 0x80000000;
	ddb_out32(IPCI_CTRLH, temp);
for(i=0;i<1000;i++);
	temp &= ~0xc0000000;
	ddb_out32(IPCI_CTRLH, temp);
}

unsigned __init int pcibios_assign_all_busses(void)
{
	/* we hope pci_auto has assigned the bus numbers to all buses */
	return 1;
}


void __init pcibios_fixup_resources(struct pci_dev *dev)
{
    int pos;
    int bases;

    printk("adjusting pci device: %s\n", dev->name);
 
    switch (dev->hdr_type) {
       case PCI_HEADER_TYPE_NORMAL: bases = 6; break;
       case PCI_HEADER_TYPE_BRIDGE: bases = 2; break;
       case PCI_HEADER_TYPE_CARDBUS: bases = 1; break;
       default: bases = 0; break;
    }
    for (pos=0; pos < bases; pos++) {
       struct resource* res = &dev->resource[pos];
       if (res->start >= IO_MEM_LOGICAL_START && res->end <= IO_MEM_LOGICAL_END) {
              res->start += IO_MEM_VIRTUAL_OFFSET;
              res->end += IO_MEM_VIRTUAL_OFFSET;
       }
       if (res->start >= IO_PORT_LOGICAL_START && res->end <= IO_PORT_LOGICAL_END) {
              res->start += IO_PORT_VIRTUAL_OFFSET;
              res->end += IO_PORT_VIRTUAL_OFFSET;
       }
    }

}

