#include <linux/types.h>
#include <linux/time.h>

#include <asm/time.h>
#include <asm/addrspace.h>
#include <asm/delay.h>
#include <asm/debug.h>
#include <asm/tcube.h>

#undef  DEBUG
#define RTC_DELAY

#define USE_GPIO
#ifdef  USE_GPIO
/*
 *   CE: INTP7#/GP57
 *   DO: U1DTR /GP40
 *   DI: U1CD  /GP41
 *   CLK:U1DSR /GP42 
 */

#define RTC_MASK_DO   0x0100
#define RTC_MASK_DI   0x0200
#define RTC_MASK_CLK  0x0400
#endif
 

void static rtc_set_ce(int val)
{
#ifdef DEBUG
  printk("rtc_set_ce(%d)\n",val);
#endif
  io_set16(VR7701_GPIO_DATA4, 0x0200, val ? 0xffff : 0);
#ifdef RTC_DELAY
  __delay(100000);
#endif
}

#ifdef USE_GPIO
void rtc_do(int val)
{
  io_set16(VR7701_GPIO_DATA3, RTC_MASK_DO, val ? 0xffff : 0); 
}

void rtc_clk(int val)
{
  io_set16(VR7701_GPIO_DATA3, RTC_MASK_CLK, val ? 0xffff : 0); 
}

int rtc_di(void)
{
  return io_in16(VR7701_GPIO_DATA3) & RTC_MASK_DI ? 1 : 0;
}

void rtc_out(unsigned short data)
{
  int i;
  unsigned short m;
#ifdef DEBUG
  printk(" rtc_out: %02x (%s,%d)\n",data,__FILE__,__LINE__);
#endif

  for(i=0,m=0x80; i<8;i++,m>>=1){
    rtc_clk(1);
    rtc_do(data & m);
    __delay(400);
    rtc_clk(0);
    __delay(400);
  }
}

unsigned short rtc_in(void)
{
  unsigned short data,m;
  int i;

  data = 0;
  for(i=0,m=0x80; i<8;i++,m>>=1){
    rtc_clk(1);
    __delay(400);
    if(rtc_di()) data |= m;
    rtc_clk(0);
    __delay(400);
  }
#ifdef DEBUG
  printk(" rtc_in : data=%04x (%s:%c)\n",data,__FILE__,__LINE__);
#endif
  return data;
}
#else
void static rtc_out(unsigned short data)
{
#ifdef DEBUG
  printk(" rtc_out: %02x\n",data);
#endif
  while(io_in8(VR7701_CSI_MODEREG) & 1);          /* wait until CSOT = 0 */
  io_out8(VR7701_CSI_MODEREG, 0xc0); 
#ifdef RTC_DELAY
  __delay(10000);
#endif
  io_out16(VR7701_CSI_SOTBREG, data);
  __delay(10000);
  while(io_in8(VR7701_CSI_MODEREG) & 1);         /* wait until CSOT = 0 */
}

unsigned static short rtc_in(void)
{
  unsigned short data;
  
  while(io_in8(VR7701_CSI_MODEREG) & 1);
  io_out8(VR7701_CSI_MODEREG, 0x80);
  __delay(10000);
  while(io_in8(VR7701_CSI_MODEREG) & 1);
#ifdef RTC_DELAY
  __delay(10000);
#endif
  data = io_in16(VR7701_CSI_SIRBREG);
#ifdef RTC_DELAY
  __delay(10000);
#endif
#ifdef DEBUG
  printk(" rtc_in : data=%04x\n",data);
#endif
  return data;
}
#endif

void static rtc_write(int adr, unsigned char data)
{
  rtc_set_ce(1);
  rtc_out((adr << 4)|0x08);
  rtc_out(data);
  rtc_set_ce(0);
}


void static rtc_write_burst(int adr, unsigned char *data, int dataLen)
{
  int i;
  rtc_set_ce(1);
  rtc_out((adr << 4)|0x00);
  for(i=0;i<dataLen;i++)
    rtc_out(data[i]);
  rtc_set_ce(0);
}

void static rtc_read_burst(int adr, unsigned char *data, int dataLen)
{
  int i;
  rtc_set_ce(1);
  rtc_out((adr << 4)|0x04);   /* burst read */

#ifndef USE_GPIO
  rtc_in(); /* dummy read */
#endif
  for(i=0;i<dataLen;i++)
    data[i] = rtc_in();
  rtc_set_ce(0);
}

#define BcdToDec(x) (((x)>>4)*10+((x)&0x0f))
#define DecToBcd(x)  (((x)/10<<4)+(x)%10)

static unsigned long
rtc_ricoh_rx5c348_get_time(void)
{	
  u8 date[7];
  unsigned int year, month, day, hour, minute, second;
  int tz;
  
  if(prom_get_int("TZ", &tz)==0)
    tz = 0;
  
  rtc_read_burst(0, date, sizeof(date));
  
  year   = BcdToDec(date[6]) + (date[5] & 0x80 ? 2000 : 1900);
  month  = BcdToDec(date[5] & 0x1f);
  day    = BcdToDec(date[4]);
  hour   = BcdToDec(date[2]);
  minute = BcdToDec(date[1]);
  second = BcdToDec(date[0]);

  printk(KERN_INFO "rtc_ricoh_rx5c348_get_time: %d/%02d/%02d %02d:%02d:%02d\n",
	 year, month, day, hour, minute, second);
  return mktime(year, month, day, hour, minute, second) - tz*3600;
}

int prom_get_int(const char *name, int *pVal);

static int 
rtc_ricoh_rx5c348_set_time(unsigned long t)
{
  u8 date[7];
  struct rtc_time tm;
  int tz;
  
  if(prom_get_int("TZ", &tz)==0)
    tz = 0;

  to_tm(t + tz*3600, &tm);
  date[0] = DecToBcd(tm.tm_sec);
  date[1] = DecToBcd(tm.tm_min);
  date[2] = DecToBcd(tm.tm_hour);
  date[4] = DecToBcd(tm.tm_mday);
  date[5] = DecToBcd(tm.tm_mon+1) + (tm.tm_year > 1999 ? 0x80 : 0);
  date[6] = DecToBcd(tm.tm_year > 1999 ? tm.tm_year-2000 : tm.tm_year-1900);
  
  rtc_write_burst(0, date, 3);
  rtc_write_burst(4, date+4, 3);

  printk(KERN_INFO "rtc_ricoh_rx5c348_set_time:t=%ld %d/%02d/%02d %02d:%02d:%02d\n",
	 t,tm.tm_year,tm.tm_mon+1,tm.tm_mday,tm.tm_hour,tm.tm_min,tm.tm_sec);
  return 0;
}

void rtc_ricoh_rx5c348_init(void)
{
  char data;
#ifndef USE_GPIO
  unsigned short ireg;
#endif
  int bUdump;

  if(prom_get_bool("uart1dump", &bUdump) && bUdump)
    return;
  
  
  io_set16(VR7701_GPIO_SEL,    0xc080,0xc080);      /* INT7->GPIO,U1SEL=CSI */
  io_set16(VR7701_GPIO_DATA4,  0x0200,0x0000);      /* set  GPIO57=0      */
  io_set16(VR7701_GPIO_IOSEL4, 0x0200,0xffff);      /* make GPIO57 output */

#ifdef USE_GPIO
  io_set16(VR7701_GPIO_SEL,    0xc080,0x4080);      /* U1SEL=GP,INT7->GPIO      */
  io_set16(VR7701_GPIO_DATA3,  0x0700,0x0000);      /* GP40=GP41=GP42=0          */ 
  io_set16(VR7701_GPIO_IOSEL3, 0x0700,0x0500);      /* output GP40,42,input GP41 */ 
#else
  /* io_reset16(VR7701_CSI_CNTREG, 0x8000, 1000, 0x0000); */
  io_out16(VR7701_CSI_CNTREG, 0x0000); 
  ireg = io_in16(VR7701_CSI_INTREG);
  io_out16(VR7701_CSI_INTREG,0);
  while(io_in8(VR7701_CSI_MODEREG) & 1);

  io_out8(VR7701_CSI_MODEREG,   0xc0);
  io_out8(VR7701_CSI_CLKSELREG, 0x14);
#endif
  rtc_read_burst(0x0e, &data, 1);
  if((data & 0x20)==0){
    rtc_write(0x0e, 0x20);
#ifdef RTC_DELAY
    __delay(10000);
#endif
  }
  
  /* set the function pointers */
  rtc_get_time = rtc_ricoh_rx5c348_get_time;
  rtc_set_time = rtc_ricoh_rx5c348_set_time;
}
