
#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/kdev_t.h>
#include <linux/types.h>
#include <linux/console.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/ide.h>
#include <linux/fs.h>		/* for ROOT_DEV */
#include <linux/ioport.h>
#include <linux/param.h>	/* for HZ */

#include <asm/bootinfo.h>
#include <asm/addrspace.h>
#include <asm/time.h>
#include <asm/bcache.h>
#include <asm/irq.h>
#include <asm/reboot.h>
#include <asm/gdb-stub.h>
#include <asm/debug.h>
#include <asm/tcube.h>

void rtc_ricoh_rx5c348_init(void);
int prom_get_macadr(const char *name,u_char *mac_addr);
int prom_get_bool(const char *name, int *pVal);
int prom_get_int(const char *name, int *pVal);

void pci_setup(void);		// KOSE.ADD

extern struct ide_ops std_ide_ops;

#define	CPU_COUNTER_FREQUENCY		133333000

static void TCube_machine_restart(char *command)
{
        static void (*back_to_prom) (void) = (void (*)(void)) 0xbfc00000;
	back_to_prom();
}

static void TCube_machine_halt(void)
{
	printk("TCube halted.\n");
	while (1);
}

static void TCube_machine_power_off(void)
{
	printk("TCube halted. Please turn off the power.\n");
	while (1);
}


static void __init TCube_time_init(void)
{
int tcube_rtc_init(void);
	unsigned long config = read_c0_config();
	unsigned long kose ;
	kose = ( config >> 28 ) & 7 ;
	kose = 5 * ( kose + 4 ) ;
	mips_hpt_frequency = ( CPU_COUNTER_FREQUENCY / (10*2) ) * kose ;

printk("KOSE mips_hpt_frequency=%ld\n",mips_hpt_frequency);

	tcube_rtc_init() ;
}

extern int setup_irq(unsigned int irq, struct irqaction *irqaction);
static void __init TCube_timer_setup(struct irqaction *irq)
{
	unsigned int count;

	/* printk("TCube_timer_setup:\n"); */
        /* we are using the cpu counter for timer interrupts */
	setup_irq(7, irq);

        /* to generate the first timer interrupt */
        count = read_c0_config();
       write_c0_compare(count + 10000);
}

extern void tcube_irq_setup(void);

#if defined(CONFIG_BLK_DEV_INITRD)
extern unsigned long __rd_start, __rd_end, initrd_start, initrd_end;
#endif 

static void chk_init_5701_reg(unsigned long addr, unsigned long data)
{
	unsigned long a = ddb_in32(addr);

	if( a != data ){
		printk("Unexpected 5701 reg : addr = %08lX, expected = %08lX, read = %08lX\n",
			addr + VR5701_IO_BASE, data, a );
	}
}

static void tcube_board_init(void)
{
/*
	0000_0000 ` 03FF_FFFF  64M  SDRAM
	1000_0000 ` 10FF_FFFF 128M  PCIW0(MEM)
	1800_0000 ` 187F_FFFF   8M  PCIW1(IO)
	1880_0000 ` 18FF_FFFF   8M  IOPCIW0(IO)
	1900_0000 ` 197F_FFFF   8M  IOPCIW1(MEM)
	1980_0000 ` 19FF_FFFF   8M  PCI-CFG
	1A00_0000 ` 1A7F_FFFF   8M  IOPCI-CFG
	1F00_0000 ` 1FFF_FFFF  16M  FROM
*/
	chk_init_5701_reg(0,0x1e00008f);
	chk_init_5701_reg(PADR_SDRAM01,0x000000aa);     // 64M
	chk_init_5701_reg(PADR_LOCALCS0,0x1f00004c);	// 16M
	chk_init_5701_reg(PADR_PCIW0,0x10000089);	// 16M
	chk_init_5701_reg(PADR_PCIW1,0x1800008d);       // 8M
	chk_init_5701_reg(PADR_IOPCIW0,0x1880008d);     // 8M
	chk_init_5701_reg(PADR_IOPCIW1,0x1900008d);     // 8M
	chk_init_5701_reg(LOCAL_CST0,0x00088622);
	chk_init_5701_reg(LOCAL_CFG,0x000f0000);
//	chk_init_5701_reg(GIU_DIR0,0xfbebffdf);
	chk_init_5701_reg(GIU_DIR0,0xF7EBFFDF);
	chk_init_5701_reg(GIU_DIR1,0x000007fa);
//	chk_init_5701_reg(GIU_FUNCSEL0,0xf1c1ffff);
	chk_init_5701_reg(GIU_FUNCSEL0,0xFFC1FFFF);
	chk_init_5701_reg(GIU_FUNCSEL1,0x000007f0);
	chk_init_5701_reg(GIU_PIO0,0x0e00002f);

	/* setup PCI windows - window0 for MEM/config, window1 for IO */
	ddb_set_pdar(PADR_PCIW0,   0x10000000, 0x08000000 , 32, 0, 1);
	ddb_set_pdar(PADR_PCIW1,   0x18000000, 0x00800000 , 32, 0, 1);
	ddb_set_pdar(PADR_IOPCIW0, 0x18800000, 0x00800000 , 32, 0, 1);
	ddb_set_pdar(PADR_IOPCIW1, 0x19000000, 0x00800000 , 32, 0, 1);

	/* ------------ reset PCI bus and BARs ----------------- */
	ddb_pci_reset_bus();

	ddb_out32(PCI_BAR_MEM01, 0x00000008);
	ddb_out32(PCI_BAR_MEM23, 0x00000008);

	ddb_out8 (PCI_MLTIM,     0x40);
//	ddb_out8 (PCI_MLTIM,     0x4);

	ddb_out32(PCI_BAR_LCS0, 0xffffffff);
	ddb_out32(PCI_BAR_LCS1, 0xffffffff);
	ddb_out32(PCI_BAR_LCS2, 0xffffffff);
	ddb_out32(PCI_BAR_LCS3, 0xffffffff);

	ddb_out32(IPCI_BAR_MEM01, 0x08);
	ddb_out32(IPCI_BAR_MEM23, 0xffffffff);

	ddb_out32(IPCI_BAR_LCS0, 0xffffffff);
	ddb_out32(IPCI_BAR_LCS1, 0xffffffff);
	ddb_out32(IPCI_BAR_LCS2, 0xffffffff);
	ddb_out32(IPCI_BAR_LCS3, 0xffffffff);

//	ddb_out32(IPCI_BAR_EPCIW0, 0xffffffff);
//	ddb_out32(IPCI_BAR_EPCIW1, 0xffffffff);
	ddb_out32(IPCI_BAR_IREG,   0xffffffff);

//	ddb_out32(DDB_BARC0, 0xffffffff);
//	ddb_out32(DDB_BARM230, 0xffffffff);
//	ddb_out32(DDB_BAR00, 0xffffffff);
//	ddb_out32(DDB_BAR10, 0xffffffff);
//	ddb_out32(DDB_BAR20, 0xffffffff);
//	ddb_out32(DDB_BAR30, 0xffffffff);
//	ddb_out32(DDB_BAR40, 0xffffffff);
//	ddb_out32(DDB_BAR50, 0xffffffff);
//	ddb_out32(DDB_BARB0, 0xffffffff);

//	ddb_out32(DDB_BARC1, 0xffffffff);
//	ddb_out32(DDB_BARM231, 0xffffffff);
//	ddb_out32(DDB_BAR01, 0xffffffff);
//	ddb_out32(DDB_BAR11, 0xffffffff);
//	ddb_out32(DDB_BAR21, 0xffffffff);
//	ddb_out32(DDB_BAR31, 0xffffffff);
//	ddb_out32(DDB_BAR41, 0xffffffff);
//	ddb_out32(DDB_BAR51, 0xffffffff);
//	ddb_out32(DDB_BARB1, 0xffffffff);

	/* 
	 * We use pci master register 0  for memory space / config space
	 * And we use register 1 for IO space.
	 * Note that for memory space, we bump up the pci base address
	 * so that we have 1:1 mapping between PCI memory and cpu physical.
	 * For PCI IO space, it starts from 0 in PCI IO space but with
	 * DDB_xx_IO_BASE in CPU physical address space.
	 */
	ddb_set_pmr(EPCI_INIT0, DDB_PCICMD_MEM, 0x10000000, DDB_PCI_ACCESS_32);
	ddb_set_pmr(EPCI_INIT1, DDB_PCICMD_IO,  0,          DDB_PCI_ACCESS_32);

	ddb_set_pmr(IPCI_INIT0, DDB_PCICMD_IO,  0x00800000, DDB_PCI_ACCESS_32);
	ddb_set_pmr(IPCI_INIT1, DDB_PCICMD_MEM, 0x19000000, DDB_PCI_ACCESS_32);

	/* PCI cross window should be set properly */
	ddb_set_pdar(PCI_BAR_IPCIW0, 0x18800000, 0x00800000 , 32, 0, 1);
	ddb_set_pdar(PCI_BAR_IPCIW1, 0x19000000, 0x00800000 , 32, 0, 1);
	ddb_set_pdar(IPCI_BAR_EPCIW0,   0x10000000, 0x08000000 , 32, 0, 1);
	ddb_set_pdar(IPCI_BAR_EPCIW1,   0x18000000, 0x00800000 , 32, 0, 1);


#if 0
#if defined(CONFIG_VGA_CONSOLE)
        conswitchp = &vga_con;

	screen_info = (struct screen_info) {
		0, 25,			/* orig-x, orig-y */
		0,			/* unused */
		0,			/* orig-video-page */
		0,			/* orig-video-mode */
		80,			/* orig-video-cols */
		0,0,0,			/* ega_ax, ega_bx, ega_cx */
		25,			/* orig-video-lines */
		1,			/* orig-video-isVGA */
		16			/* orig-video-points */
	};
#elif defined(CONFIG_DUMMY_CONSOLE)
        conswitchp = &dummy_con;
#endif
#endif

}

extern struct pci_controller ddb5477_ext_controller;
extern struct pci_controller ddb5477_io_controller;

//void __init shima_tcube_setup(void)
static int shima_tcube_setup(void)
{
	extern int panic_timeout;

	/* printk("TCube_setup %s:%d\n",__FILE__,__LINE__); */
//	irq_setup = tcube_irq_setup;
	set_io_port_base(0xB8000000);

	board_time_init    = TCube_time_init;
	board_timer_setup  = TCube_timer_setup;

	_machine_restart   = TCube_machine_restart;
	_machine_halt      = TCube_machine_halt;
	_machine_power_off = TCube_machine_power_off;

#ifdef CONFIG_BLK_DEV_IDE
//KOSE041014	ide_ops = &std_ide_ops;
#endif

	/* setup resource limits */
	ioport_resource.end = 0x01000000;
	iomem_resource.end  = 0xffffffff;
	
	/* Reboot on panic */
	panic_timeout = 180;

#ifdef CONFIG_FB
//	conswitchp = &dummy_con;
#endif

	tcube_board_init();

#if 0
#if defined(CONFIG_BLK_DEV_INITRD)
	ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
	initrd_start = (unsigned long)&__rd_start;
	initrd_end =   (unsigned long)&__rd_end;
#endif
#endif
	panic_timeout = 180;

printk("KOSE register_pci_controller 1\n");
	register_pci_controller (&ddb5477_ext_controller);
printk("KOSE register_pci_controller 2\n");
	register_pci_controller (&ddb5477_io_controller);
printk("KOSE register_pci_controller 3\n");

	return 0;
}

early_initcall(shima_tcube_setup);

void __init bus_error_init(void)
{
  /* do nothing */
}

