/*
 * RTC DRIVER for TCube
 */

#include <linux/config.h>

#include <linux/fs.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/module.h>
 
#include <asm/io.h>
//#include <asm/rtc.h>
#include <asm/time.h>
#include <asm/uaccess.h>
#include <asm/tcube.h>

#include <linux/miscdevice.h>

//#define RTC_MAJOR_NR 121 /* local major, change later */
#define RTC_MAJOR_NR 135 /* local major, change later */

#include "ricoh_5c348.c"

#undef BCD_TO_BIN
#define BCD_TO_BIN(val) (((val)&15) + ((val)>>4)*10)

#undef BIN_TO_BCD
#define BIN_TO_BCD(val) ((((val)/10)<<4) + (val)%10)

static const char ricoh_5c348_name[] = "ricoh_5c348";

static unsigned long rtc_tcube_get_time(void)
{
    unsigned char ss[10] ;
    unsigned int year, month, day, hour, minute, second;
    ReadClock( ss );
    year    = BCD_TO_BIN(ss[0] )+ ((ss[1] & 0x80) ? 2000 : 1900 );
    month   = BCD_TO_BIN(ss[1] & ~0x80);
    day     = BCD_TO_BIN(ss[2] );
    hour    = BCD_TO_BIN(ss[3] );
    minute  = BCD_TO_BIN(ss[4] );
    second  = BCD_TO_BIN(ss[5] );
    return mktime(year, month, day, hour, minute, second);
}

static int rtc_tcube_set_time(unsigned long t)
{
    struct rtc_time tm;
    unsigned char ss[10];
    /* convert */
    to_tm(t, &tm);
    ss[0] = BIN_TO_BCD(tm.tm_year % 100 );
    ss[1] = BIN_TO_BCD(tm.tm_mon ) + ((tm.tm_year >= 2000) ? 0x80 : 0 );
    ss[2] = BIN_TO_BCD(tm.tm_mday) ;
    ss[3] = BIN_TO_BCD(tm.tm_hour) ;
    ss[4] = BIN_TO_BCD(tm.tm_min) ;
    ss[5] = BIN_TO_BCD(tm.tm_sec) ;
    SetClock( ss );
}

static int rtc_ioctl(
  struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg) 
{

    struct rtc_time rtc_tm;
    unsigned char ss[10];

    switch(cmd) {
        case RTC_RD_TIME:   /* read the time/date from RTC  */
        {
            ReadClock( ss );
            rtc_tm.tm_sec  = BCD_TO_BIN(ss[5] );
            rtc_tm.tm_min  = BCD_TO_BIN(ss[4] );
            rtc_tm.tm_hour = BCD_TO_BIN(ss[3] );
            rtc_tm.tm_mday = BCD_TO_BIN(ss[2] );
            rtc_tm.tm_mon  = BCD_TO_BIN(ss[1] & ~0x80);
            rtc_tm.tm_year = BCD_TO_BIN(ss[0] )+ ((ss[1] & 0x80) ? 2000 : 1900 );
            if (copy_to_user((struct rtc_time*)arg, &rtc_tm, sizeof(struct rtc_time)))
                return -EFAULT; 
            return 0;
        }

        case RTC_SET_TIME:  /* set the RTC */
        {
            if (copy_from_user(&rtc_tm, (struct rtc_time*)arg, sizeof(struct rtc_time)))
                return -EFAULT;     
            ss[0] = BIN_TO_BCD(rtc_tm.tm_year % 100 );
            ss[1] = BIN_TO_BCD(rtc_tm.tm_mon ) + ((rtc_tm.tm_year >= 2000) ? 0x80 : 0) ;
            ss[2] = BIN_TO_BCD(rtc_tm.tm_mday) ;
            ss[3] = BIN_TO_BCD(rtc_tm.tm_hour) ;
            ss[4] = BIN_TO_BCD(rtc_tm.tm_min) ;
            ss[5] = BIN_TO_BCD(rtc_tm.tm_sec) ;
            SetClock( ss );
            return 0;
        }
        default:
            printk("rtc_ioctl Error:%lX\n", cmd );
            return -ENOIOCTLCMD;
    }
}

static void tt_func(){}


static struct file_operations rtc_fops = {
        .owner      = THIS_MODULE,
        .ioctl      = rtc_ioctl,
/*		.llseek		= tt_func,
		.read		= tt_func,
		.poll		= tt_func,
		.open		= tt_func,
		.release	= tt_func,*/
}; 

static struct miscdevice ricoh_dev=
{
	.minor	= RTC_MINOR,
	.name	= "rtc",
	.fops	= &rtc_fops,
};

int tcube_rtc_init(void) 
{ 
	rtc_get_time = rtc_tcube_get_time;
	rtc_set_time = rtc_tcube_set_time;

#ifdef KOSE________________________________________________________
	if (register_chrdev(RTC_MAJOR_NR, ricoh_5c348_name, &rtc_fops)) {
		printk(KERN_INFO "%s: unable to get major %d for rtc\n", 
		       ricoh_5c348_name, RTC_MAJOR_NR);
		return -1;
	}else{
		printk("%s: registered\n", ricoh_5c348_name );
	}
#endif
	return 0;
}

void tcube_rtc_init_misc(void) 
{
	if(misc_register(&ricoh_dev)){
		printk("%s: register Error\n", ricoh_5c348_name );
		return -1;
	}else{
		printk("%s: registered\n", ricoh_5c348_name );
	}
}
