/*
 * LynxEM+/EM4+(Silicon Motion Inc.) fb driver
 *						for VR41xx/VR5xxx.
 *
 * Copyright (C) NEC Corporation 2001, 2002.
 * Author: t-miwa@pi.jp.nec.co.jp
 *
 * THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
 * ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
 * PARTICULAR PURPOSE.
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/selection.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/console.h>

#include "../console/fbcon.h"
//#include <video/fbcon.h>
//#include <video/fbcon-mfb.h>
//#include <video/fbcon-cfb2.h>
//#include <video/fbcon-cfb4.h>
#if 0
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>
#endif

#include "smifb.h"
#include "smi_hw.h"
//#include "smi_params.h"


#define DPPORT_BASE_OFFSET		0x400000
#define DP_BASE_OFFSET			0x408000
#define VP_BASE_OFFSET			0x40c000
#define CP_BASE_OFFSET			0x40e000
#define IO_BASE_OFFSET			0x700000


#define DPPORT_REGION_SIZE		(32*1024)
#define DPREG_REGION_SIZE		(16*1024)
#define VPREG_REGION_SIZE		(8*1024)
#define CPREG_REGION_SIZE		(8*1024)
#define MMIO_REGION_SIZE		(1*1024*1024)


#define LYNX3DM_DPPORT_BASE_OFFSET		0x100000
#define LYNX3DM_DP_BASE_OFFSET			0x000000
#define LYNX3DM_VP_BASE_OFFSET			0x000800
#define LYNX3DM_CP_BASE_OFFSET			0x001000
#define LYNX3DM_IO_BASE_OFFSET			0x0c0000
#define LYNX3DM_FB_BASE_OFFSET			0x200000


#define LYNX3DM_DPPORT_REGION_SIZE		(1024*1024)
#define LYNX3DM_DPREG_REGION_SIZE		(2*1024)
#define LYNX3DM_VPREG_REGION_SIZE		(2*1024)
#define LYNX3DM_CPREG_REGION_SIZE		(2*1024)
#define LYNX3DM_MMIO_REGION_SIZE		(256*1024)

/*
 * Card Identification
 *
 */
static struct pci_device_id smifb_pci_tbl[] __devinitdata = {
	{PCI_VENDOR_ID_SMI, PCI_DEVICE_ID_SMI_LYNX_EM_PLUS,
	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},	/* Lynx EM+/EM4+ */
	{PCI_VENDOR_ID_SMI, PCI_DEVICE_ID_SMI_LYNX_3DM,
	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},	/* Lynx 3DM/3DM+/3DM4+ */
	{0, } /* terminate list */
};
MODULE_DEVICE_TABLE(pci, smifb_pci_tbl);

/*
 *
 * global variables
 *
 */
static char noaccel __initdata = 1;
#ifndef MODULE
static const char *mode_option __initdata = NULL;
#endif


#if 0
/* 1024x768, 8bpp, 60Hz */
static struct fb_var_screeninfo smifb_default_var = {
	xres:				1024,
	yres:				768,
	xres_virtual:		1024,
	yres_virtual:		768,
	xoffset:			0,
	yoffset:			0,
	bits_per_pixel: 	8,
	grayscale:			0,
	red:				{0, 6, 0},
	green:				{0, 6, 0},
	blue:				{0, 6, 0},
	transp: 			{0, 0, 0},
	nonstd: 			0,
	activate:			0,
	height: 			-1,
	width:				-1,
	accel_flags:		0,
	pixclock:			39721, /* D */
	left_margin:		138,
	right_margin:		24,
	upper_margin:		24,
	lower_margin:		4,
	hsync_len:			160,
	vsync_len:			6,
	sync:				0,
	vmode:				FB_VMODE_NONINTERLACED
};
#endif
#if 1
/* 1024x768, 16bpp, 60Hz */
static struct fb_var_screeninfo smifb_default_var = {
	xres:				1024,
	yres:				768,
	xres_virtual:		1024,
	yres_virtual:		768,
	xoffset:			0,
	yoffset:			0,
	bits_per_pixel: 	16,
	grayscale:			0,
	red:				{11, 5, 0},
	green:				{ 5, 6, 0},
	blue:				{ 0, 5, 0},
	transp: 			{ 0, 0, 0},
	nonstd: 			0,
	activate:			0,
	height: 			-1,
	width:				-1,
	accel_flags:		0,
	pixclock:			39721, /* D */
	left_margin:		138,
	right_margin:		24,
	upper_margin:		24,
	lower_margin:		4,
	hsync_len:			160,
	vsync_len:			6,
	sync:				0,
	vmode:				FB_VMODE_NONINTERLACED
};
#endif
#if 0
/* 800x600, 8bpp, 60Hz */
static struct fb_var_screeninfo smifb_default_var = {
	xres:				800,
	yres:				600,
	xres_virtual:		800,
	yres_virtual:		600,
	xoffset:			0,
	yoffset:			0,
	bits_per_pixel: 	8,
	grayscale:			0,
	red:				{0, 6, 0},
	green:				{0, 6, 0},
	blue:				{0, 6, 0},
	transp: 			{0, 0, 0},
	nonstd: 			0,
	activate:			0,
	height: 			-1,
	width:				-1,
	accel_flags:		0,
	pixclock:			39721, /* D */
	left_margin:		82,
	right_margin:		16,
	upper_margin:		19,
	lower_margin:		1,
	hsync_len:			152,
	vsync_len:			4,
	sync:				0,
	vmode:				FB_VMODE_NONINTERLACED
};
#endif
#if 0
/* 800x600, 24bpp, 60Hz */
static struct fb_var_screeninfo smifb_default_var = {
	xres:				800,
	yres:				600,
	xres_virtual:		800,
	yres_virtual:		600,
	xoffset:			0,
	yoffset:			0,
	bits_per_pixel: 	24,
	grayscale:			0,
	red:				{16, 8, 0},
	green:				{8, 8, 0},
	blue:				{0, 8, 0},
	transp: 			{0, 0, 0},
	nonstd: 			0,
	activate:			0,
	height: 			-1,
	width:				-1,
	accel_flags:		0,
	pixclock:			39721, /* D */
	left_margin:		82,
	right_margin:		16,
	upper_margin:		19,
	lower_margin:		1,
	hsync_len:			152,
	vsync_len:			4,
	sync:				0,
	vmode:				FB_VMODE_NONINTERLACED
};
#endif

#if 0
/* 400x232, 24bpp, 60Hz */
static struct fb_var_screeninfo smifb_default_var = {
	xres:				400,
	yres:				232,
	xres_virtual:		400,
	yres_virtual:		232,
	xoffset:			0,
	yoffset:			0,
	bits_per_pixel: 	16,
	grayscale:			0,
	red:				{11, 5, 0},
	green:				{ 5, 6, 0},
	blue:				{ 0, 5, 0},
	transp: 			{ 0, 0, 0},
	nonstd: 			0,
	activate:			0,
	height: 			-1,
	width:				-1,
	accel_flags:		0,
	pixclock:			100000, /* D */
	left_margin:		82,
	right_margin:		16,
	upper_margin:		19,
	lower_margin:		1,
	hsync_len:			152,
	vsync_len:			4,
	sync:				0,
	vmode:				FB_VMODE_NONINTERLACED
};
#endif


static char drvrname[] = "NEC video driver for SMI LynxEM+";

/*
 *
 * general utility functions
 *
 */
PRIVATE void
smi_set_dispsw(struct smifb_info *sinfo, struct display *disp)
{
#if 0
	struct fb_info *info= &sinfo->info;
//	int accel = disp->var.accel_flags & FB_ACCELF_TEXT;
	int accel = info->var.accel_flags & FB_ACCELF_TEXT;


	SMI_DBG("smi_set_dispsw");

	disp->dispsw_data = NULL;
	disp->screen_base = sinfo->fb_base;
	disp->type = FB_TYPE_PACKED_PIXELS;
	disp->type_aux = 0;
	disp->ypanstep = 1;
	disp->ywrapstep = 0;
	disp->can_soft_blank = 1;
	disp->inverse = 0;

	switch(disp->var.bits_per_pixel){
#ifdef FBCON_HAS_CFB8
	case 8:
		/* sinfo->dispsw = accel ? fbcon_smi8 : fbcon_cfb8; */
		sinfo->dispsw = fbcon_cfb8;
		disp->dispsw = &sinfo->dispsw;
		disp->line_length = disp->var.xres_virtual;
		disp->visual = FB_VISUAL_PSEUDOCOLOR;
		break;
#endif
#ifdef FBCON_HAS_CFB16
	case 16:
		/* sinfo->dispsw = accel ? fbcon_smi16 : fbcon_cfb16; */
		sinfo->dispsw = fbcon_cfb16;
		disp->dispsw_data = &sinfo->con_cmap.cfb16;
		disp->dispsw = &sinfo->dispsw;
		disp->line_length = disp->var.xres_virtual * 2;
		//disp->visual = FB_VISUAL_DIRECTCOLOR;
		disp->visual = FB_VISUAL_TRUECOLOR;
		break;
#endif
#ifdef FBCON_HAS_CFB24
	case 24:
		/* sinfo->dispsw = accel ? fbcon_smi24 : fbcon_cfb24; */
		sinfo->dispsw = fbcon_cfb24;
		disp->dispsw_data = &sinfo->con_cmap.cfb24;
		disp->dispsw = &sinfo->dispsw;
		disp->line_length = disp->var.xres_virtual * 3;
		//disp->visual = FB_VISUAL_DIRECTCOLOR;
		disp->visual = FB_VISUAL_TRUECOLOR;
		break;
#endif
#ifdef FBCON_HAS_CFB32
	case 32:
		/* sinfo->dispsw = accel ? fbcon_smi32 : fbcon_cfb32; */
		sinfo->dispsw = fbcon_cfb32;
		disp->dispsw_data = &sinfo->con_cmap.cfb32;
		disp->dispsw = &sinfo->dispsw;
		disp->line_length = disp->var.xres_virtual * 4;
		//disp->visual = FB_VISUAL_DIRECTCOLOR;
		disp->visual = FB_VISUAL_TRUECOLOR;
		break;
#endif
	default:
		SMI_MSG("Setting fbcon_dummy renderer\n");
		sinfo->dispsw = fbcon_dummy;
		disp->dispsw = &sinfo->dispsw;
	};

/*
	if(sinfo->cursor){
		sinfo->dispsw.cursor = smifb_cursor;
		sinfo->dispsw.set_font = smifb_set_font;
	}
*/
#endif
}


PRIVATE void
smi_load_video_mode(struct smifb_info *sinfo, struct fb_var_screeninfo *video_mode)
{
	int bpp, width, height;
	int hDisplaySize, hDisplay, hStart, hEnd, hTotal;
	int vDisplay, vStart, vEnd, vTotal;
	int dotClock;

	SMI_DBG("smi_load_video_mode");

	SMI_MSG("smi_load_video_mode: video_mode->xres = %d\n", video_mode->xres);
	SMI_MSG("smi_load_video_mode:             yres = %d\n", video_mode->yres);
	SMI_MSG("smi_load_video_mode:             xres_virtual = %d\n", video_mode->xres_virtual);
	SMI_MSG("smi_load_video_mode:             yres_virtual = %d\n", video_mode->yres_virtual);
	SMI_MSG("smi_load_video_mode:             xoffset = %d\n", video_mode->xoffset);
	SMI_MSG("smi_load_video_mode:             yoffset = %d\n", video_mode->yoffset);
	SMI_MSG("smi_load_video_mode:             bits_per_pixel = %d\n", video_mode->bits_per_pixel);

	/* smifb_blank(1, (struct fb_info*)sinfo); */

	bpp = video_mode->bits_per_pixel;
	if(bpp == 16 && video_mode->green.length == 5)
		bpp = 15;

	/* horizontal params */
	width		 = video_mode->xres_virtual;
	hDisplaySize = video_mode->xres;									/* number of pixels for one horizontal line */
	hDisplay	 = (hDisplaySize /8) -1;								/* number of character clocks */
	hStart		 = (hDisplaySize + video_mode->right_margin)/8 +2;		/* h-blank start character clocks */
	hEnd		 = (hDisplaySize + video_mode->right_margin
					+ video_mode->hsync_len) / 8 - 1;					/* h-sync end */
	hTotal		 = (hDisplaySize + video_mode->right_margin
					+ video_mode->hsync_len
					+ video_mode->left_margin) / 8 - 1; 				/* character clock from h-sync to next h-sync */

	/* vertical params */
	height	 = video_mode->yres_virtual;
	vDisplay = video_mode->yres - 1;									/* number of lines */
	vStart	 = video_mode->yres + video_mode->lower_margin - 1; 		/* v-sync pulse start*/
	vEnd	 = video_mode->yres + video_mode->lower_margin
				+ video_mode->vsync_len - 1;							/* v-sync end */
	vTotal	 = video_mode->yres + video_mode->lower_margin
				+ video_mode->vsync_len + video_mode->upper_margin + 2; /* number of scanlines (v-blank end) */

	dotClock = 1000000000 / video_mode->pixclock;


	smi_set_moderegs(sinfo, bpp, width, height,
					 hDisplaySize,
					 hDisplay, hStart, hEnd, hTotal,
					 vDisplay, vStart, vEnd, vTotal,
					 dotClock, video_mode->sync);
}


/*
 * internal fb_ops helper functions
 *
 */
PRIVATE int
smi_get_cmap_len(const struct fb_var_screeninfo *var)
{
	int rc;
#if 0

	//SMI_DBG("smi_get_cmap_len");

	switch(var->bits_per_pixel){
#ifdef FBCON_HAS_CFB8
	case 8:
		rc = 256;		/* pseudocolor... 256 entries HW palette */
		break;
#endif
#ifdef FBCON_HAS_CFB16
	case 16:
		/* fall through */
#endif
#ifdef FBCON_HAS_CFB24
	case 24:
		/* fall through */
#endif
#ifdef FBCON_HAS_CFB32
	case 32:
		/* fall through */
#endif
	default:
		rc = 16;		/* directcolor... 16 entries SW palette */
		break;
	};

#endif
	return rc;
}

PRIVATE int
smi_getcolreg(unsigned regno, unsigned *red, unsigned *green,
			  unsigned *blue, unsigned *transp,
			  struct fb_info *info)
{
#if 0
	struct smifb_info *sinfo = (struct smifb_info *)info;

	//SMI_DBG("smi_getcolreg");

//	if(regno >= smi_get_cmap_len(&sinfo->currcon_display->var))
	if(regno >= smi_get_cmap_len(&sinfo->info->var))
		return 1;


	*red	= (sinfo->palette[regno].red   << 8) | sinfo->palette[regno].red ;
	*green	= (sinfo->palette[regno].green << 8) | sinfo->palette[regno].green;
	*blue	= (sinfo->palette[regno].blue  << 8) | sinfo->palette[regno].blue;
	*transp = 0;

//	  SMI_MSG(" regno = % 4d: r,g,b: 0x%08x, 0x%08x, 0x%08x\n", regno, *red, *green, *blue);

#endif
	return 0;
}

PRIVATE int
smi_setcolreg(unsigned regno, unsigned red, unsigned green,
			  unsigned blue, unsigned transp,
			  struct fb_info *info)
{
#if 0
	struct smifb_info *sinfo = (struct smifb_info *)info;
	struct display *dsp;

	//SMI_DBG("smi_setcolreg");


	dsp = sinfo->currcon_display;
//	if(regno >= smi_get_cmap_len(&dsp->var))
	if(regno >= smi_get_cmap_len(&sinfo->info->var))
		return -EINVAL;

	sinfo->palette[regno].red	= red >> 8;
	sinfo->palette[regno].green = green >> 8;
	sinfo->palette[regno].blue	= blue >> 8;

	if(dsp->var.grayscale){
		/* gray = 0.30*R + 0.59*G + 0.11*B */
		red = green = blue =
			(red * 77 + green * 151 + blue * 28) >> 10;
	}

//	  SMI_MSG(" regno = % 4d: r,g,b: 0x%04x (0x%02x), 0x%04x (0x%02x), 0x%04x (0x%02x)\n",
//			regno, red, (u8)(red>>10), green, (u8)(green>>10), blue, (u8)(blue>>10));
	switch(dsp->var.bits_per_pixel){
#ifdef FBCON_HAS_CFB8
	case 8:
		/* "transparent" stuff is completely ignored. */
		VGA_WRITE8(sinfo->mmio, 0x3c6, 0xff);			/* DAC address mask */
		VGA_WRITE8(sinfo->mmio, 0x3c8, (u8)regno);		/* DAC address write */
		VGA_WRITE8(sinfo->mmio, 0x3c9, (u8)(red >> 10));
		VGA_WRITE8(sinfo->mmio, 0x3c9, (u8)(green >> 10));
		VGA_WRITE8(sinfo->mmio, 0x3c9, (u8)(blue >> 10));
		break;
#endif
#ifdef FBCON_HAS_CFB16
	case 16:
		if(dsp->var.green.length == 5){
			/* R:G:B == 5:5:5 */
			sinfo->con_cmap.cfb16[regno] =
				((red & 0xf800) >> 1) | ((green & 0xf800) >> 6) | ((blue & 0xf800) >> 11);
		}
		else{
			/* R:G:B == 5:6:5 */
			sinfo->con_cmap.cfb16[regno] =
				((red & 0xf800) >> 0) | ((green & 0xfc00) >> 5) | ((blue & 0xf800) >> 11);
		}
		break;
#endif
#ifdef FBCON_HAS_CFB24
	case 24:
		sinfo->con_cmap.cfb24[regno] = 
			((red & 0xff00) << 8) | ((green & 0xff00)) | ((blue & 0xff00) >> 8);
		break;
#endif
#ifdef FBCON_HAS_CFB32
	case 32:
		sinfo->con_cmap.cfb32[regno] =
			((red & 0xff00) << 8) | ((green & 0xff00)) | ((blue & 0xff00) >> 8);
		break;
#endif
	};

#endif
	return 0;
}


/*
 *
 * framebuffer operations
 *
 */
PRIVATE int
smifb_get_fix(struct fb_fix_screeninfo *fix, int con, struct fb_info *info)
{
	struct smifb_info *sinfo = (struct smifb_info *)info;
	struct display *dsp;

	SMI_DBG("smifb_get_fix");

//	dsp = (con < 0) ? sinfo->info.disp : &fb_display[con];

//	memset(fix, 0, sizeof(struct fb_fix_screeninfo));

	fix->smem_start = sinfo->fb_base_phys;
	fix->smem_len	= sinfo->fbsize;
	fix->mmio_start = sinfo->dpr_base_phys;
	fix->mmio_len	= sinfo->dpport_size;

//	fix->type	  = dsp->type;
//	fix->type_aux = dsp->type_aux;
//	fix->visual   = dsp->visual;

	fix->xpanstep  = 0; /* FIXME: no xpanstep for now */
	fix->ypanstep  = 1; /* FIXME: no ypanstep for now */
	fix->ywrapstep = 0; /* FIXME: no ywrap for now */

//	fix->line_length = dsp->line_length;

	// fix->accel = FB_ACCEL_SM712;

	return 0;
}

PRIVATE int
smifb_get_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
#if 0
	struct smifb_info *sinfo = (struct smifb_info *)info;

	SMI_DBG("smifb_get_var");

	*var = (con < 0)? sinfo->disp.var : fb_display[con].var;

#endif
	return 0;
}

PRIVATE int
smifb_set_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
#if 0
	struct smifb_info *smiinfo = (struct smifb_info *)info;
	struct display *dsp;
	struct fb_var_screeninfo v;
	int nom, den;
	int accel;
	unsigned chgvar = 0;

	SMI_DBG("smifb_set_var");

	SMI_MSG("con = %d\n", con);
	SMI_MSG("Requested: % 4d x % 4d x % 2d\n", var->xres, var->yres, var->bits_per_pixel);
	SMI_MSG("  virtual: % 4d x % 4d\n", var->xres_virtual, var->yres_virtual);
	SMI_MSG("   offset: (%d, %d)\n", var->xoffset, var->yoffset);
	SMI_MSG("grayscale: %d\n", var->grayscale);


	dsp = (con < 0) ? smiinfo->info.disp : & fb_display[con];

	/* if var has changed, we shoud call changevar() later */
	if(con >= 0){
		chgvar = ((dsp->var.xres != var->xres) ||
				  (dsp->var.yres != var->yres) ||
				  (dsp->var.xres_virtual != var->xres_virtual) ||
				  (dsp->var.yres_virtual != var->yres_virtual) ||
				  (dsp->var.accel_flags != var->accel_flags) ||
				  (dsp->var.bits_per_pixel != var->bits_per_pixel)
				  || memcmp(&dsp->var.red, &var->red,
							sizeof(var->red))
				  || memcmp(&dsp->var.green, &var->green,
							sizeof(var->green))
				  || memcmp(&dsp->var.blue, &var->blue,
							sizeof(var->blue)));
	}

	memcpy(&v, var, sizeof(v));

	accel = v.accel_flags & FB_ACCELF_TEXT;

	switch(v.bits_per_pixel){
#ifdef FBCON_HAS_CFB8
	case 1 ... 8:
		v.bits_per_pixel = 8;
		nom = 1;
		den = 1;
		v.red.offset = 0;
		v.red.length = 8;
		v.green.offset = 0;
		v.green.length = 8;
		v.blue.offset = 0;
		v.blue.length = 8;
		break;
#endif /* FBCON_HAS_CFB8 */
#ifdef FBCON_HAS_CFB16
	case 9 ... 15:
		v.green.length = 5;
		/* fall through */
	case 16:
		v.bits_per_pixel = 16;
		nom = 2;
		den = 1;
		if (v.green.length == 5) {
			/* 0rrrrrgg gggbbbbb */
			v.red.offset = 10;
			v.green.offset = 5;
			v.blue.offset = 0;
			v.red.length = 5;
			v.green.length = 5;
			v.blue.length = 5;
		} else {
			/* rrrrrggg gggbbbbb */
			v.red.offset = 11;
			v.green.offset = 5;
			v.blue.offset = 0;
			v.red.length = 5;
			v.green.length = 6;
			v.blue.length = 5;
		}
		break;
#endif /* FBCON_HAS_CFB16 */
#ifdef FBCON_HAS_CFB24
	case 17 ... 24:
		v.bits_per_pixel = 24;
		nom = 3;
		den = 1;
		v.red.offset = 16;
		v.green.offset = 8;
		v.blue.offset = 0;
		v.red.length = 8;
		v.green.length = 8;
		v.blue.length = 8;
		break;
#endif /* FBCON_HAS_CFB24 */
#ifdef FBCON_HAS_CFB32
	case 25 ... 32:
		v.bits_per_pixel = 32;
		nom = 4;
		den = 1;
		v.red.offset = 16;
		v.green.offset = 8;
		v.blue.offset = 0;
		v.red.length = 8;
		v.green.length = 8;
		v.blue.length = 8;
		break;
#endif /* FBCON_HAS_CFB32 */
	default:
		SMI_MSG("mode %dx%dx%d rejected...color depth not supported.\n",
			   var->xres, var->yres, var->bits_per_pixel);
		SMI_MSG("EXIT, returning -EINVAL\n");
		return -EINVAL;
	};

	memcpy(&dsp->var, &v, sizeof(v));
	if (chgvar) {
		smi_set_dispsw(smiinfo, dsp);

		if (accel) {
			/*
			if (nomove)
				dsp->scrollmode = SCROLL_YNOMOVE;
			else
				dsp->scrollmode = 0;
			*/
		}
		else{
			dsp->scrollmode = SCROLL_YREDRAW;
		}

//		if (info && info->changevar)
//			info->changevar(con);
	}

	/* smifb_creat_cursor() */
	smi_load_video_mode(smiinfo, &v);
	/*
	if(accel)
		smi_setup_accel();
	*/

#endif
	return 0;
}

PRIVATE int
smifb_get_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
#if 0
	struct smifb_info *sinfo = (struct smifb_info *)info;
	struct display *dsp;

	SMI_DBG("smifb_get_cmap");

	dsp = (con < 0) ? sinfo->info.disp : &fb_display[con];

	if(con == sinfo->currcon){	/* current console? */
		int rc = fb_get_cmap(cmap, kspc, smi_getcolreg, info);
		return rc;
	}
	else if(dsp->cmap.len){ 	/* non default colormap? */
		fb_copy_cmap(&dsp->cmap, cmap, kspc ? 0 : 2);
	}
	else{
		fb_copy_cmap(fb_default_cmap
					 (smi_get_cmap_len(&dsp->var)), cmap,
					 kspc ? 0 : 2);
	}

#endif
	return 0;
}

PRIVATE int
smifb_set_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
#if 0
	struct smifb_info *sinfo = (struct smifb_info *)info;
	struct display *dsp;
	unsigned int cmap_len;

	SMI_DBG("smifb_set_cmap");

	dsp = (con < 0) ? sinfo->info.disp: &fb_display[con];

	cmap_len = smi_get_cmap_len(&dsp->var);
	if(dsp->cmap.len != cmap_len){
		int err = fb_alloc_cmap(&dsp->cmap, cmap_len, 0);
		if(err){
			return err;
		}
	}
	if(con == sinfo->currcon){	/* current console? */
		int rc = fb_set_cmap(cmap, kspc, smi_setcolreg, info);
		return rc;
	}
	else{
		fb_copy_cmap(cmap, &dsp->cmap, kspc?0:1);
	}

#endif
	return 0;
}


PRIVATE int
smifb_pan_display(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
#if 0
	SMI_DBG("smifb_pan_display");

	/* use the CRT08 ??? */
#endif
	return 0;
}

PRIVATE int
smifb_ioctl(struct inode *inode, struct file *file, u_int cmd,
			u_long arg, int con, struct fb_info *info)
{
#if 0
	SMI_DBG("smifb_ioctl");

	/* no Lynx-specific ioctls */
#endif
	return -EINVAL;
}


/* tell fb to switch consoles */
PRIVATE int
smifb_switch(int con, struct fb_info *info)
{
#if 0
	struct smifb_info *sinfo = (struct smifb_info *)info;
	struct fb_cmap *cmap;
	struct display *dsp;

	SMI_DBG("smifb_switch");

	dsp = (con < 0) ? sinfo->info.disp : &fb_display[con];

	if(sinfo->currcon >= 0){
		/* Do we have to save the colormap? */
		cmap = &(sinfo->currcon_display->cmap);
		SMI_MSG("switch1: con = %d, cmap.len = %d\n",
				sinfo->currcon, cmap->len);
		if(cmap->len){
			SMI_MSG("switch1a: %p %p %p %p\n", cmap->red, cmap->green, cmap->blue, cmap->transp);
			fb_get_cmap(cmap, 1, smi_getcolreg, info);
		}
	}
	sinfo->currcon = con;
	sinfo->currcon_display = dsp;

//	smifb_set_var(&dsp->var, con, info);
	smifb_set_var(&sinfo->info->var, con, info);
	smi_set_dispsw(sinfo, dsp);

#endif
	return 0;
}


/* tell fb to update the vars */
PRIVATE int
smifb_updatevar(int con, struct fb_info *info)
{
#if 0
	int rc;

	SMI_DBG("smifb_updatevar");

	rc = (con < 0) ? -EINVAL : smifb_pan_display(&fb_display[con].var,
												 con, info);
	return rc;
#endif
}


/*
 * tell fb to (un)blank the screen
 *		arg = 0: unblank
 *		arg > 0: VESA level (arg-1)
 */
void
smifb_blank(int con, struct fb_info* info)
{
	SMI_DBG("smifb_blank");
	return;
}

static int vgxfb_setcolreg(unsigned regno, unsigned red, unsigned green, unsigned blue, unsigned transp, struct fb_info *info)
{
	if (regno > 15)
		return 1;

	((u16 *)(info->pseudo_palette))[regno] = (red & 0xf800) | (green & 0xfc00 >> 5) | (blue & 0xf800 >> 11);
	return 0;
}

/*
 * Initialization helper functions
 *
 */
/* kernel interface */
static struct fb_ops smifb_ops = {
#if 0
	owner:		THIS_MODULE,
	fb_get_fix: 		smifb_get_fix,
	fb_get_var: 		smifb_get_var,
	fb_set_var: 		smifb_set_var,
	fb_get_cmap:		smifb_get_cmap,
	fb_set_cmap:		smifb_set_cmap,
	fb_pan_display: 	smifb_pan_display,
	fb_ioctl:			smifb_ioctl,
//	  fb_mmap:			smifb_mmap,
//	  fb_rasterimg: 	smifb_rasterimg,
#endif
	.owner        =	THIS_MODULE,
	.fb_setcolreg = vgxfb_setcolreg,
	.fb_fillrect  = cfb_fillrect,
	.fb_copyarea  = cfb_copyarea,
	.fb_imageblit = cfb_imageblit,
	.fb_cursor	= soft_cursor,
};

PRIVATE int __devinit
smi_init_disp_var(struct smifb_info *sinfo)
{
#if 0
#ifndef MODULE
	if(mode_option)
//		fb_find_mode(&sinfo->disp.var, &sinfo->info, mode_option, NULL, 0, NULL, 8);
		fb_find_mode(&sinfo->info.var, &sinfo->info, mode_option, NULL, 0, NULL, 8);
#endif /* not MODULE */
#endif
	return 0;
}

PRIVATE int __devinit
smi_init_disp(struct smifb_info *sinfo)
{
#if 0
	struct fb_info *info;
	struct display *disp;

	SMI_DBG("smi_init_disp");

	info = &sinfo->info;
	disp = &sinfo->disp;

//	disp->var = smifb_default_var;
	info->var = smifb_default_var;

	if(noaccel)
//		disp->var.accel_flags &= ~FB_ACCELF_TEXT;
		info->var.accel_flags &= ~FB_ACCELF_TEXT;
	else
//		disp->var.accel_flags |= FB_ACCELF_TEXT;
		info->var.accel_flags |= FB_ACCELF_TEXT;

	info->disp = disp;

	sinfo->currcon_display = disp;

	if(smi_init_disp_var(sinfo) < 0){
		return -1;
	}

	smi_set_dispsw(sinfo, disp);

#endif
	return 0;
}


PRIVATE int __devinit
smi_set_fbinfo(struct smifb_info *sinfo)
{
#if 0
	struct fb_info *info;

	SMI_DBG("smi_set_fbinfo");

	info = &(sinfo->info);
//	strcpy(info->modename, sinfo->drvr_name);

	info->node = -1;
	info->flags = FBINFO_FLAG_DEFAULT;
	info->fbops = &smifb_ops;

//	info->display_fg = NULL;
//	info->changevar = NULL;

//	info->switch_con = &smifb_switch;

//	info->updatevar = &smifb_updatevar;
//	info->blank = &smifb_blank;

	if(smi_init_disp(sinfo) < 0) /* must be done last */
		return -1;
#endif
	return 0;
}


/*
 * VGA registers
 *
 */
PRIVATE void
Unlock(struct smifb_info *sinfo)
{
	SMI_DBG("Unlock");
	regSR_write(sinfo->mmio, 0x33, regSR_read(sinfo->mmio, 0x33) & 0x20);
}

PRIVATE void
Lock(struct smifb_info *sinfo)
{
	SMI_DBG("Lock");
}

PRIVATE void
UnlockVGA(struct smifb_info *sinfo)
{
	SMI_DBG("UnlockVGA");
	regCR_write(sinfo->mmio, 0x11, regCR_read(sinfo->mmio, 0x11) & 0x7f);
}

PRIVATE void
LockVGA(struct smifb_info *sinfo)
{
	SMI_DBG("LockVGA");
	regCR_write(sinfo->mmio, 0x11, regCR_read(sinfo->mmio, 0x11) | 0x80);
}


static struct fb_fix_screeninfo vgxfb_fix = {
	.id 		= "vgxFB",
//	.smem_start 	= VGA_START,
//	.smem_len 	= VGA_SIZE,
	.type 		= FB_TYPE_PACKED_PIXELS,
	.visual 	= FB_VISUAL_TRUECOLOR,
//	.line_length 	= 640*2,
	.line_length 	= 1024*2,
	.accel 		= FB_ACCEL_NONE,
};

static u32 colreg[17];

/*
 * PCI bus
 *
 */
PRIVATE int __devinit
smifb_probe( struct pci_dev *pd, const struct pci_device_id *ent )
{
	int len;
	int res;
	u16 cmd;
	struct smifb_info *sinfo;
	struct fb_info *info;

	SMI_DBG("smifb_probe");

//	SMI_MSG("driver name      %s\n", pd->name);

	SMI_MSG("vendor id        %04x\n", pd->vendor);
	SMI_MSG("device id        %04x\n", pd->device);
	SMI_MSG("sub vendor id    %04x\n", pd->subsystem_vendor);
	SMI_MSG("sub device id    %04x\n", pd->subsystem_device);

	SMI_MSG("base0 start addr %08x\n", (unsigned int)pci_resource_start(pd, 0));
	SMI_MSG("base0 end   addr %08x\n", (unsigned int)pci_resource_end(pd, 0));
	SMI_MSG("base0 region len %08x\n", (unsigned int)pci_resource_len(pd, 0));
	SMI_MSG("base0 flags      %08x\n", (unsigned int)pci_resource_flags(pd, 0));

	pci_read_config_word(pd, PCI_STATUS, &cmd);
	SMI_MSG("PCI status      %04x\n", cmd);

	pci_read_config_word(pd, PCI_COMMAND, &cmd);
	SMI_MSG("PCI command      %04x\n", cmd);

	cmd |= PCI_COMMAND_MEMORY|PCI_COMMAND_IO;
	pci_write_config_word(pd, PCI_COMMAND, cmd);

	pci_read_config_word(pd, PCI_STATUS, &cmd);
	SMI_MSG("PCI status      %04x\n", cmd);
	pci_read_config_word(pd, PCI_COMMAND, &cmd);
	SMI_MSG("PCI command      %04x\n", cmd);


	/* allocate memory resources */
	sinfo = kmalloc(sizeof(struct smifb_info), GFP_KERNEL);
	if(!sinfo){
		goto err_out;
	}
	memset(sinfo, 0, sizeof(struct smifb_info));

	/* driver name */
	sinfo->drvr_name = drvrname;

	sinfo->pd = pd;
	sinfo->base_phys = pci_resource_start(sinfo->pd, 0);		/* Frame Buffer base address */
	len = pci_resource_len(sinfo->pd, 0);
//	  SMI_MSG("len = %d\n", len);
	SMI_MSG("len = %lX\n", len);
	if(!request_mem_region(sinfo->base_phys, len, "smifb")){
		printk(KERN_ERR "cannot reserve FrameBuffer and MMIO region\n");
		goto err_out_kfree;
	}

	if((res = pci_enable_device(sinfo->pd)) < 0){
		printk("smifb: failed to enable -- err %d\n", res);
		goto err_out_free_base;
	}

	pci_read_config_word(pd, PCI_COMMAND, &cmd);
	SMI_MSG("PCI command      %04x\n", cmd);


	{
		unsigned int pseudo_io, pseudo_io_len;
		unsigned char *pseudo_io_p;

#ifndef CONFIG_SHIMA_TCUBE
		pseudo_io = pci_resource_start(sinfo->pd, 1);
		pseudo_io_len = pci_resource_len(sinfo->pd, 1);
		
		pseudo_io_p = ioremap(pseudo_io, pseudo_io_len);
#else
		*(unsigned long *)0xbe000610 = 0x10000012 ; // CHANGE to PCI IO ACCESS
		asm("sync");
		pseudo_io = pci_resource_start(sinfo->pd, 0);
		pseudo_io_len = pci_resource_len(sinfo->pd, 0);
		pseudo_io_p = ioremap(pseudo_io, pseudo_io_len);
//		pseudo_io_p = 0xb4000000 ;
#endif

		VGA_WRITE8(pseudo_io_p, 0x3c2, 0x2b); /* Miscellaneous Output Register ( write 0x3c2, read 0x3cc ) */
		VGA_WRITE8(pseudo_io_p, 0x3c3, 0x40); /* */
		regSR_write(pseudo_io_p, 0x00, 0x00);
		regSR_write(pseudo_io_p, 0x17, 0xe2);
		regSR_write(pseudo_io_p, 0x18, 0xff);
		
#ifndef CONFIG_SHIMA_TCUBE
		iounmap(pseudo_io_p);
#else
		iounmap(pseudo_io_p);
		*(unsigned long *)0xbe000610 = 0x10000016 ; // ɖ߂(PCI MEM ACCESS)
		asm("sync");
#endif
	}
	sinfo->base = ioremap(sinfo->base_phys, len); /* FB+DPD+DPR+VPR+CPR+MMIO */
	if(!sinfo->base){
		goto err_out_free_base;
	}
	switch((sinfo->pd)->device){
	case PCI_DEVICE_ID_SMI_LYNX_EM_PLUS:
		sinfo->dpport = (caddr_t)(sinfo->base + DPPORT_BASE_OFFSET);
		sinfo->dpr	  = (caddr_t)(sinfo->base + DP_BASE_OFFSET);
		sinfo->vpr	  = (caddr_t)(sinfo->base + VP_BASE_OFFSET);
		sinfo->cpr	  = (caddr_t)(sinfo->base + CP_BASE_OFFSET);
#if defined(CONFIG_MIPS_ITE8172) || defined(CONFIG_NEC_EAGLE) || defined(CONFIG_NEC_HAWK)
		sinfo->mmio   = (caddr_t)(sinfo->base + 0x05000000);
#else
		sinfo->mmio   = (caddr_t)(sinfo->base + IO_BASE_OFFSET);
#endif
		sinfo->fb_base = (caddr_t)(sinfo->base + 0);
		break;
	case PCI_DEVICE_ID_SMI_LYNX_3DM:
		sinfo->dpport = (caddr_t)(sinfo->base + LYNX3DM_DPPORT_BASE_OFFSET);
		sinfo->dpr	  = (caddr_t)(sinfo->base + LYNX3DM_DP_BASE_OFFSET);
		sinfo->vpr	  = (caddr_t)(sinfo->base + LYNX3DM_VP_BASE_OFFSET);
		sinfo->cpr	  = (caddr_t)(sinfo->base + LYNX3DM_CP_BASE_OFFSET);
#if defined(CONFIG_MIPS_ITE8172) || defined(CONFIG_NEC_EAGLE) || defined(CONFIG_NEC_HAWK)
		sinfo->mmio   = (caddr_t)(sinfo->base + 0x05000000);
#else
		sinfo->mmio   = (caddr_t)(sinfo->base + LYNX3DM_IO_BASE_OFFSET);
#endif
		sinfo->fb_base = (caddr_t)(sinfo->base + LYNX3DM_FB_BASE_OFFSET);
		break;
	}
//	  regSR_write(sinfo->mmio, 0x17, 0x22);
	regSR_write(sinfo->mmio, 0x18, 0x11);

	SMI_MSG("sinfo->dpport = 0x%08x\n", (u_int32_t)sinfo->dpport);
	SMI_MSG("sinfo->dpr  = 0x%08x, sinfo->vpr   = 0x%08x\n",
			(unsigned int)sinfo->dpr, (unsigned int)sinfo->vpr);
	SMI_MSG("sinfo->cpr  = 0x%08x, sinfo->mmio  = 0x%08x\n",
			(unsigned int)sinfo->cpr, (unsigned int)sinfo->mmio);

	/* Set the chip in color mode and unlock the registers */
	VGA_WRITE8(sinfo->mmio, 0x3c2, 0x2b/*0x67*/); /* Miscellaneous Output Register ( write 0x3c2, read 0x3cc ) */

	Unlock(sinfo);
	UnlockVGA(sinfo);

	/* save the current chip status */

	/* Enable linearmemory */
#if 0
	regSR_write(sinfo->mmio, 0x17, regSR_read(sinfo->mmio, 0x17) & ~0x02);
	regSR_write(sinfo->mmio, 0x18, regSR_read(sinfo->mmio, 0x18) | 0x19);
#endif
	switch((sinfo->pd)->device){
	case PCI_DEVICE_ID_SMI_LYNX_EM_PLUS:
		regSR_write(sinfo->mmio, 0x62, 0xff);
		regSR_write(sinfo->mmio, 0x6a, 0x0c);
		regSR_write(sinfo->mmio, 0x6b, 0x02);

		*(u32 *)(sinfo->fb_base +4) = 0xaa551133;
		SMI_MSG("       *(u32 *)(sinfo->fb_base +4) = 0x%08x\n",		*(u32 *)(sinfo->fb_base +4));
		if(*(u32 *)(sinfo->fb_base +4) != 0xaa551133){
			/* Program the MCLK to 130MHz */
			regSR_write(sinfo->mmio, 0x6a, 0x10);
			regSR_write(sinfo->mmio, 0x6b, 0x02);
			regSR_write(sinfo->mmio, 0x62, 0x3e);
			sinfo->fbsize = 2*1024*1024;		/* LynxEM+ */
			SMI_MSG("ChipID = LynxEM+. Force the MCLK to 85MHz and the memory size to 2MiB\n");
		}
		else{
			sinfo->fbsize = 4*1024*1024;		/* LynxEM4+ */
			SMI_MSG("ChipID = LynxEM4+. Force the MCLK to 85MHz and the memory size to 4MiB\n");
		}
		sinfo->fb_base_phys = sinfo->base_phys;
		break;
	case PCI_DEVICE_ID_SMI_LYNX_3DM:
		{
			int tmp;
			int mem_table[4] = { 8, 16, 0, 4 };
			tmp = (regSR_read(sinfo->mmio, 0x76) & 0xff);
			SMI_MSG("%02x\n", tmp);
			sinfo->fbsize = mem_table[(tmp >> 6)] * 1024 * 1024;
			
			regSR_write(sinfo->mmio, 0x62, 0xff);
			regSR_write(sinfo->mmio, 0x6a, 0x0c);
			regSR_write(sinfo->mmio, 0x6b, 0x02);
			
			sinfo->fb_base_phys = sinfo->base_phys + LYNX3DM_FB_BASE_OFFSET;
		}
		break;
	default:
		/* this driver supports only LynxEM+/EM4+ */
		goto err_out_free_base;
	};

#if 0
	/* setup HW Cursor (if need) */
	if(smi_set_fbinfo(sinfo) < 0){
		goto err_out_free_base;
	}
#endif

	info = &(sinfo->info);

	smifb_get_fix(&vgxfb_fix, -1, info);

	info->flags = FBINFO_FLAG_DEFAULT;
	info->fbops = &smifb_ops;
	info->var = smifb_default_var;
	info->fix = vgxfb_fix;
	info->pseudo_palette = colreg;
	info->screen_base = sinfo->fb_base;
printk("KOSE info->screen_base=%08lX\n", info->screen_base);

	smi_load_video_mode(sinfo, &smifb_default_var);

	if(register_framebuffer(&sinfo->info) < 0){
		goto err_out_free_base;
	}
	pci_set_drvdata(pd, sinfo);

	printk(KERN_INFO "smifb: " "framebuffer (%s)\n",
//	printk(KERN_INFO "smifb: " "%s framebuffer (%s)\n",
//		   sinfo->pd->name,
		   sinfo->drvr_name);
	return 0;

// err_out_load_state: /* restore initail chip status */
 err_out_free_base:
	release_mem_region(sinfo->base_phys, len);
 err_out_kfree:
	kfree(sinfo);
 err_out:
	return -ENODEV;
}

PRIVATE void __devexit
smifb_remove( struct pci_dev *pd )
{
	struct smifb_info *sinfo = pci_get_drvdata(pd);
	SMI_DBG("smifb_remove");

	if(!sinfo)
		return;

	unregister_framebuffer(&sinfo->info);

	/* stop the lynx chip */

	release_mem_region(sinfo->base_phys, pci_resource_len(sinfo->pd, 0));
	kfree(sinfo);
	pci_set_drvdata(pd, NULL);
}


/*
 * Initialization
 *
 */
#ifndef MODULE
int __init
smifb_setup(char *options)
{
	SMI_DBG("smifb_setup");

	if(!options || options)
		return 0;

	return 0;
}
#endif /* not MODULE */


static struct pci_driver smifb_driver = {
	.name=		 "smifb",
	.id_table=	 smifb_pci_tbl,
	.probe= 	 smifb_probe,
	.remove=	 __devexit_p(smifb_remove),
};


/*
 * Driver initialization
 */
int __init
smifb_init(void)
{
	SMI_DBG("smifb_init");

	return pci_module_init(&smifb_driver);
}

/*
 * modularization
 *
 */
#ifdef MODULE
PRIVATE void __exit smifb_exit(void)
{
	pci_unregister_driver(&smifb_driver);
}

module_init(smifb_init);
module_exit(smifb_exit);

#endif /* MODULE */

module_init(smifb_init);				// KOSE.ADD

MODULE_AUTHOR("Takeshi Tsukada");
MODULE_DESCRIPTION("Framebuffer driver for Silicon Motion, Inc. Lynx 3DM+");
MODULE_LICENSE("GPL");
