#define FBCON_HAS_CFB8
#define FBCON_HAS_CFB16
#define FBCON_HAS_CFB24
/*
 * LynxEM+/EM4+(Silicon Motion Inc.) fb driver
 *						for VR41xx/VR5xxx.
 *
 * Copyright (C) NEC Corporation 2001, 2002.
 * Author: t-miwa@pi.jp.nec.co.jp
 *
 * THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
 * ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
 * PARTICULAR PURPOSE.
 *
 *
 * Lynx H/W control routines.
 *
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>


#include <linux/fb.h>
#ifdef KOSE______________________________________________________________________
#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb2.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>
#endif

#include "smifb.h"
#include "smi_hw.h"
#include "smi_params.h"


struct smi_mode_regs ModeDefaultParams = {
	DISPLAY_1024x768x8,
	/* reg_MISC */
	0xEB,
	{	/* reg_SR00_SR04 */
		0x03, 0x01, 0x0F, 0x03, 0x0E,
	},
	{	/* reg_SR10_SR24 */
		0xF3, 0xB6, 0xC0, 0xDD, 0x00, 0x0E, 0x17, 0xe2,
		0xff, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xC4, 0x30, 0x02, 0x01, 0x01,
	},
	{	/* reg_SR30_SR75 */
		0x38, 0x03, 0x20, 0x09, 0xC0, 0x3A, 0x3A, 0x3A,
		0x3A, 0x3A, 0x3A, 0x3A, 0x00, 0x00, 0x03, 0xFF,
		0x00, 0xFC, 0x00, 0x00, 0x20, 0x18, 0x00, 0xFC,
		0x20, 0x0C, 0x44, 0x20, 0x00, 0x3A, 0x3A, 0x3A,
		0x06, 0x68, 0xA7, 0x7F, 0x83, 0x24, 0xFF, 0x03,
		0x00, 0x60, 0x59, 0x3A, 0x3A, 0x00, 0x00, 0x3A,
		0x01, 0x80, 0x7E, 0x1A, 0x1A, 0x00, 0x00, 0x00,
		0x50, 0x03, 0x74, 0x14, 0x12, 0x82, 0x09, 0x02,
		0x04, 0x45, 0x30, 0x30, 0x40, 0x20,
	},
	{	/* reg_SR80_SR93 */
		0xFF, 0x07, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x3A,
		0xF7, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x3A, 0x3A,
		0x00, 0x00, 0x00, 0x00,
	},
	{	/* reg_SRA0_SRAF */
		0x00, 0xFB, 0x9F, 0x01, 0x00, 0xED, 0xED, 0xED,
		0x7B, 0xFB, 0xFF, 0xFF, 0x97, 0xEF, 0xBF, 0xDF,
	},
	{	/* reg_GR00_GR08 */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F,
		0xFF,
	},
	{	/* reg_AR00_AR14 */
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x41, 0x00, 0x0F, 0x00, 0x00,
	},
	{	/* reg_CR00_CR18 */
		0xA3, 0x7F, 0x7F, 0x00, 0x85, 0x16, 0x24, 0xF5,
		0x00, 0x60, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x03, 0x09, 0xFF, 0x80, 0x40, 0xFF, 0x00, 0xE3,
		0xFF,
	},
	{	/* reg_CR30_CR4D */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x02, 0x20,
		0x00, 0x00, 0x00, 0x40, 0x00, 0xFF, 0xBF, 0xFF,
		0xA3, 0x7F, 0x00, 0x86, 0x15, 0x24, 0xFF, 0x00,
		0x01, 0x07, 0xE5, 0x20, 0x7F, 0xFF,
	},
	{	/* reg_CR90_CRA7 */
		0x55, 0xD9, 0x5D, 0xE1, 0x86, 0x1B, 0x8E, 0x26,
		0xDA, 0x8D, 0xDE, 0x94, 0x00, 0x00, 0x18, 0x00,
		0x03, 0x03, 0x03, 0x03, 0x03, 0x03, 0x15, 0x03,
	},
};



/*
 * set mode registers
 */
void
smi_set_moderegs(struct smifb_info *sinfo,
				 int bpp, int width, int height,
				 int hDisplaySize,
				 int hDisplay, int hStart, int hEnd, int hTotal,
				 int vDisplay, int vStart, int vEnd, int vTotal,
				 int dotClock, int sync)
{
	int i;
	int tmp_mode = SMI_DEFAULT_MODE;
	int lineLength;
//	  int ScreenWidth, ScreenStride, ScreenHeight, ScreenV;
	struct smi_mode_regs curMode;

	SMI_DBG("smi_set_moderegs");

	SMI_MSG("bpp = %d, width = %d, height = %d\n", bpp, width, height);
	SMI_MSG("hDisplaySize = %d\n", hDisplaySize);
	SMI_MSG("hDisplay = %d, hStart = %d, hEnd = %d, hTotal = %d\n",
			hDisplay, hStart, hEnd, hTotal);
	SMI_MSG("vDisplay = %d, vStart = %d, vEnd = %d, vTotal = %d\n",
			vDisplay, vStart, vEnd, vTotal);
	SMI_MSG("dotClock = %d\n", dotClock);

	lineLength = width * bpp / 8;

	switch(bpp){
#ifdef FBCON_HAS_CFB8
	case 8:
SMI_MSG("smi_set_moderegs.bpp=8\n");
		if(hDisplaySize <= 640)
			tmp_mode = DISPLAY_640x480x8;
		else if(width <= 800)
			tmp_mode = DISPLAY_800x600x8;
		else if(width <= 1024)
			tmp_mode = DISPLAY_1024x768x8;
		else if(width <= 1280)
			tmp_mode = DISPLAY_1280x1024x8;
		reg_DPR10(sinfo) = (lineLength << 16) | lineLength; 	/* RowPitch */
		reg_DPR1E(sinfo) = 0x0005;
		reg_DPR3C(sinfo) = (lineLength << 16) | lineLength; 	/* Dst & Src Window Width */
		reg_VPR00(sinfo) = 0x0 << 16;
		break;
#endif
#ifdef FBCON_HAS_CFB16
	case 16:
SMI_MSG("smi_set_moderegs.bpp=16\n");
		if(hDisplaySize <= 400)
			tmp_mode = DISPLAY_LCD_400x232x16;
		if(hDisplaySize <= 640)
			tmp_mode = DISPLAY_640x480x16;
		else if(width <= 800)
			tmp_mode = DISPLAY_800x600x16;
		else if(width <= 1024)
			tmp_mode = DISPLAY_1024x768x16;
		reg_DPR10(sinfo) = (lineLength/2 << 16) | lineLength/2; /* RowPitch */
		reg_DPR1E(sinfo) = 0x0015;
		reg_DPR3C(sinfo) = (lineLength/2 << 16) | lineLength/2; /* Dst & Src Window Width */
		reg_VPR00(sinfo) = 0x2 << 16;
		break;
#endif
#ifdef FBCON_HAS_CFB24
	case 24:
SMI_MSG("smi_set_moderegs.bpp=24\n");
		if(hDisplaySize <= 640)
			tmp_mode = DISPLAY_640x480x24;
		else if(width <= 800)
			tmp_mode = DISPLAY_800x600x24;
		reg_DPR10(sinfo) = (lineLength/3 << 16) | lineLength/3; /* RowPitch */
		reg_DPR1E(sinfo) = 0x0035;
		reg_DPR3C(sinfo) = (lineLength/3 << 16) | lineLength/3; /* Dst & Src Window Width */
		reg_VPR00(sinfo) = 0x4 << 16;
		break;
#endif
	};

#if 1					// default:=0
	for(i = 0; i < modeNums; i++){
		if(ModeInitParams[i].mode == tmp_mode)
			break;
	}
	if(i == modeNums)
		tmp_mode = SMI_DEFAULT_MODE;
#else
	tmp_mode = DISPLAY_LCD_400x232x16;
#endif

	memcpy(&curMode, &ModeInitParams[tmp_mode], sizeof(struct smi_mode_regs));

#if 1
	/*
	 * Override some Mode Params
	 */
	/* MISC Reg */
	curMode.reg_MISC = 0x30 | (hDisplay == 640) ? 0x03 : 0x0b;
	if(sync & FB_SYNC_HOR_HIGH_ACT)
		curMode.reg_MISC |= 0x40;
	if(sync & FB_SYNC_VERT_HIGH_ACT)
		curMode.reg_MISC |= 0x80;

	/* CRTC */
	curMode.reg_CR00_CR18[0x00] = (u8)(hTotal - 4);
	curMode.reg_CR00_CR18[0x01] = (u8)hDisplay;
	curMode.reg_CR00_CR18[0x02] = (u8)hDisplay;
	curMode.reg_CR00_CR18[0x03] = 0x00;
	curMode.reg_CR00_CR18[0x04] = (u8)hStart;
	curMode.reg_CR00_CR18[0x05] = (hEnd & 0x1f);
	curMode.reg_CR00_CR18[0x06] = (u8)(vTotal & 0xff);
	curMode.reg_CR00_CR18[0x07] = (u8)(((vStart >> 9) & 0x01) << 7)
		| (u8)(((vDisplay >> 9) & 0x01) << 6)
		| (u8)(((vTotal   >> 9) & 0x01) << 5)
		| 1 << 4							   /* D (LC) */
		| (u8)(((vStart   >> 8) & 0x01) << 2)
		| (u8)(((vDisplay >> 8) & 0x01) << 1)
		| (u8)((vTotal	  >> 8) & 0x01);

	curMode.reg_CR00_CR18[0x09] = (u8)(vDisplay >> 9) << 5
		| 1 << 6;												/* D (LC bit9) */
	curMode.reg_CR00_CR18[0x10] = (u8)(vStart & 0xff);
	curMode.reg_CR00_CR18[0x11] = (u8)(vEnd & 0xf);
	curMode.reg_CR00_CR18[0x12] = (u8)(vDisplay & 0xff);
	curMode.reg_CR00_CR18[0x13] = ((width / 8) * ((bpp + 1) / 8)) & 0xFF;
	curMode.reg_CR00_CR18[0x15] = (u8)(vDisplay & 0xff);
	curMode.reg_CR00_CR18[0x16] = 0x00;

	curMode.reg_CR00_CR18[0x14] = (hDisplaySize > 1024)? 0x00: 0x40; /* D */ /* Underline Location */

	/* Extended CRTC */
	curMode.reg_CR30_CR4D[0x30-0x30] = (u8)(((vTotal >> 10) & 0x01) << 3)
		| (u8)(((vDisplay >> 10) & 0x01) << 1)
		| (u8)((vStart >> 10) & 0x1);					/* D (CRTD) (CVDER)*/
#endif


curMode.reg_SR30_SR75[0x32] = 0xff; //KOSE. (Memory Type and Timig Control Reg)

	
	/* Set VGA Mode */
	for(i = 0; i <= SIZE_SR00_SR04; i++)
		regSR_write(sinfo->mmio, 0x00+i, curMode.reg_SR00_SR04[i]);
	for(i = 0; i <= SIZE_SR10_SR24; i++)
		regSR_write(sinfo->mmio, 0x10+i, curMode.reg_SR10_SR24[i]);
	for(i = 0; i <= SIZE_SR30_SR75; i++)
		regSR_write(sinfo->mmio, 0x30+i, curMode.reg_SR30_SR75[i]);
	for(i = 0; i <= SIZE_SR80_SR93; i++)
		regSR_write(sinfo->mmio, 0x80+i, curMode.reg_SR80_SR93[i]);
	for(i = 0; i <= SIZE_SRA0_SRAF; i++)
		regSR_write(sinfo->mmio, 0xA0+i, curMode.reg_SRA0_SRAF[i]);
	for(i = 0; i <= SIZE_GR00_GR08; i++)
		regGR_write(sinfo->mmio, 0x00+i, curMode.reg_GR00_GR08[i]);
	for(i = 0; i <= SIZE_AR00_AR14; i++)
		regAR_write(sinfo->mmio, 0x00+i, curMode.reg_AR00_AR14[i]);
	for(i = 0; i <= SIZE_CR00_CR18; i++)
		regCR_write(sinfo->mmio, 0x00+i, curMode.reg_CR00_CR18[i]);
	for(i = 0; i <= SIZE_CR30_CR4D; i++)
		regCR_write(sinfo->mmio, 0x30+i, curMode.reg_CR30_CR4D[i]);
	for(i = 0; i <= SIZE_CR90_CRA7; i++)
		regCR_write(sinfo->mmio, 0x90+i, curMode.reg_CR90_CRA7[i]);

	/* SetMemoryMapRegisters */
	reg_DPR14(sinfo) = 0xffffffff;		/* FG color */
	reg_DPR18(sinfo) = 0x00000000;		/* BG color */
	reg_DPR24(sinfo) = 0xffffffff;		/* Color Mask */
	reg_DPR28(sinfo) = 0xffff;			/* Masks */
	reg_DPR2C(sinfo) = 0;
	reg_DPR30(sinfo) = 0;
	reg_DPR34(sinfo) = 0xffffffff;
	reg_DPR38(sinfo) = 0xffffffff;
	reg_DPR40(sinfo) = 0;
	reg_DPR44(sinfo) = 0;

	reg_VPR0C(sinfo) = 0;

	reg_VPR10(sinfo) = ((lineLength/8+2)<<16) | (lineLength/8);

	reg_VPR40(sinfo) = 0;

	reg_VPR28(sinfo) = 0x00000000;
	reg_VPR2C(sinfo) = ((hDisplaySize-1)<<16) | (vDisplay);
	reg_VPR30(sinfo) = 0x00000000;
	reg_VPR34(sinfo) = (lineLength << 16) | lineLength;
}

